"""Randomized iterative methods for decompositions.
"""
from numbers import Integral

import numpy as np
import scipy.linalg as sla
from cytoolz import identity

from ..gen.rand import randn
from ..core import dag, dot, njit


def lu_orthog(X):
    return sla.lu(X, permute_l=True, overwrite_a=True, check_finite=False)[0]


def qr_orthog(X):
    return sla.qr(X, mode='economic', overwrite_a=True, check_finite=False)[0]


def orthog(X, lu=False):
    if lu:
        return lu_orthog(X)
    return qr_orthog(X)


def QB_to_svd(Q, B, compute_uv=True):
    UsV = sla.svd(B, full_matrices=False, compute_uv=compute_uv,
                  overwrite_a=True, check_finite=False)

    if not compute_uv:
        return UsV

    U, s, V = UsV
    return dot(Q, U), s, V


def trim(arrays, k):
    if isinstance(arrays, tuple) and len(arrays) == 3:
        U, s, VH = arrays
        U, s, VH = U[:, :k], s[:k], VH[:k, :]
        return U, s, VH
    if isinstance(arrays, tuple) and len(arrays) == 2:
        # Q, B factors
        Q, B = arrays
        return Q[:, :k], B[:k, :]
    else:
        # just singular values
        return arrays[:k]


def possibly_extend_randn(G, k, p, A):
    # make sure we are using block of the right size by removing or adding
    kG = G.shape[1]
    if kG > k + p:
        # have too many columns
        G = G[:, :k + p]
    elif kG < k + p:
        # have too few columns
        G_extra = randn((A.shape[1], k + p - kG), dtype=A.dtype)
        G = np.concatenate((G, G_extra), axis=1)
    return G


def isstring(x, s):
    if not isinstance(x, str):
        return False
    return x == s


def rsvd_qb(A, k, q, p, state, AH=None):

    if AH is None:
        AH = dag(A)

    # generate first block
    if isstring(state, 'begin-qb'):
        G = randn((A.shape[1], k + p), dtype=A.dtype)
    # block already supplied
    elif len(state) == 1:
        G, = state
    # mid-way through adaptive algorithm in QB mode
    if len(state) == 3:
        Q, B, G = state
    else:
        Q = np.empty((A.shape[0], 0), dtype=A.dtype)
        B = np.empty((0, A.shape[1]), dtype=A.dtype)

    QH, BH = dag(Q), dag(B)
    G = possibly_extend_randn(G, k, p, A)

    Qi = orthog(dot(A, G) - dot(Q, dot(B, G)), lu=q > 0)

    for i in range(1, q + 1):
        Qi = orthog(dot(AH, Qi) - dot(BH, dot(QH, Qi)), lu=True)
        Qi = orthog(dot(A, Qi) - dot(Q, dot(B, Qi)), lu=i != q)

    Qi = orthog(Qi - dot(Q, dot(QH, Qi)))
    Bi = dag(dot(AH, Qi)) - dot(dot(dag(Qi), Q), B)

    if p > 0:
        Qi, Bi = trim((Qi, Bi), k)

    Q = np.concatenate((Q, Qi), axis=1)
    B = np.concatenate((B, Bi), axis=0)

    return Q, B, G


def rsvd_core(A, k, compute_uv=True, q=2, p=0, state=None, AH=None):
    """Core R3SVD algorithm.

    Parameters
    ----------
    A : linear operator, shape (m, n)
        Operator to decompose, assumed m >= n.
    k : int
        Number of singular values to find.
    compute_uv : bool, optional
        Return the left and right singular vectors.
    q : int, optional
        Number of power iterations.
    p : int, optional
        Over sampling factor.
    state : {None, array_like, (), (G0,), (U0, s0, VH0, G0)}, optional
        Iterate based on these previous results:

            - None: basic mode.
            - array_like: use this as the initial subspace.
            - 'begin-svd': begin block iterations, return U, s, VH, G
            - (G0,) : begin block iterations with this subspace
            - (U0, s0, VH0, G0): continue block iterations, return G

    """
    iterating = isinstance(state, (tuple, str))
    maybe_project_left = maybe_project_right = identity

    if AH is None:
        AH = dag(A)

    # generate first block
    if state is None or isstring(state, 'begin-svd'):
        G = randn((A.shape[1], k + p), dtype=A.dtype)
    # initial block supplied
    elif hasattr(state, 'shape'):
        G = state
    elif len(state) == 1:
        G, = state
    # mid-way through adaptive algorithm in SVD mode
    elif len(state) == 4:
        U0, s0, VH0, G = state
        UH0, V0 = dag(U0), dag(VH0)

        def maybe_project_left(X):
            X -= dot(U0, dot(UH0, X))
            return X

        def maybe_project_right(X):
            X -= dot(V0, dot(VH0, X))
            return X

    G = possibly_extend_randn(G, k, p, A)
    G = maybe_project_right(G)

    Q = dot(A, G)
    Q = maybe_project_left(Q)
    Q = orthog(Q, lu=q > 0)

    # power iterations with stabilization
    for i in range(1, q + 1):
        Q = dot(AH, Q)
        Q = maybe_project_right(Q)
        Q = orthog(Q, lu=True)

        Q = dot(A, Q)
        Q = maybe_project_left(Q)
        Q = orthog(Q, lu=i < q)

    B = dag(dot(AH, Q))
    UsVH = QB_to_svd(Q, B, compute_uv=compute_uv or iterating)
    if p > 0:
        UsVH = trim(UsVH, k)

    if not iterating:
        return UsVH

    U, s, VH = UsVH

    if isstring(state, 'begin-svd') or len(state) == 1:
        # first run -> don't need to project or concatenate anything
        return U, s, VH, G

    U = orthog(maybe_project_left(U))
    VH = dag(orthog(maybe_project_right(dag(VH))))

    U = np.concatenate((U0, U), axis=1)
    s = np.concatenate((s0, s))
    VH = np.concatenate((VH0, VH), axis=0)

    return U, s, VH, G


@njit
def is_sorted(x):  # pragma: no cover
    for i in range(x.size - 1):
        if x[i + 1] < x[i]:
            return False
    return True


def gen_k_steps(start, incr=1.4):
    yield start
    step = start
    while True:
        yield step
        step = round(incr * step)


def rsvd_iterate(A, eps, compute_uv=True, q=2, p=0, G0=None,
                 k_max=None, k_start=2, k_incr=1.4, AH=None, use_qb=20):
    """Handle rank-adaptively calling ``rsvd_core``.
    """

    if AH is None:
        AH = dag(A)

    # perform first iteration and set initial rank
    k_steps = gen_k_steps(k_start, k_incr)
    rank = next(k_steps)

    if use_qb:
        Q, B, G = rsvd_qb(A, rank, q=q, p=p, AH=AH,
                          state='begin-qb' if G0 is None else (G0,))
        U, s, VH = QB_to_svd(Q, B)
        G -= dot(dag(VH), dot(VH, G))
    else:
        U, s, VH, G = rsvd_core(A, rank, q=q, p=p, AH=AH,
                                state='begin-svd' if G0 is None else (G0,))

    # perform randomized SVD in small blocks
    while (s[-1] > eps * s[0]) and (rank < k_max):

        # only step k as far as k_max
        new_k = min(next(k_steps), k_max - rank)
        rank += new_k

        if (rank < use_qb) or (use_qb is True):
            Q, B, G = rsvd_qb(A, new_k, q=q, p=p, state=(Q, B, G), AH=AH)
            U, s, VH = QB_to_svd(Q, B)
            G -= dot(dag(VH), dot(VH, G))
        else:
            # concatenate new U, s, VH orthogonal to current U, s, VH
            U, s, VH, G = rsvd_core(A, new_k, q=q, p=p,
                                    state=(U, s, VH, G), AH=AH)

    # make sure singular values always sorted in decreasing order
    if not is_sorted(s):
        so = np.argsort(s)[::-1]
        U, s, VH = U[:, so], s[so], VH[so, :]

    return U, s, VH if compute_uv else s


@njit
def count_svdvals_needed(s, eps):  # pragma: no cover
    n = s.size
    thresh = eps * s[0]
    for i in range(n - 1, 0, -1):
        if s[i - 1] < thresh:
            n -= 1
        else:
            break
    return n


def isdouble(dtype):
    """Check if ``dtype`` is double precision.
    """
    return dtype in ('float64', 'complex128')


def estimate_rank(A, eps, k_max=None, use_sli=True, k_start=2, k_incr=1.4,
                  q=0, p=0, get_vectors=False, G0=None, AH=None, use_qb=20):
    """Estimate the rank of an linear operator. Uses a low quality random
    SVD with a resolution of ~ 10.

    Parameters
    ----------
    A : linear operator
        The operator to find rank of.
    eps : float
        Find rank to this relative (compared to largest singular value)
        precision.
    k_max : int, optional
        The maximum rank to find.
    use_sli : bool, optional
        Whether to use :func:`scipy.linalg.interpolative.estimate_rank` if
        possible (double precision and no ``k_max`` set).
    k_start : int, optional
        Begin the adaptive SVD with a block of this size.
    k_incr : float, optional
        Adaptive rank increment factor. Increase the k-step (from k_start) by
        this factor each time. Set to 1 to use a constant step.
    q : int, optional
        Number of power iterations.
    get_vectors : bool, optional
        Return the right singular vectors found in the pass.
    G0 : , optional

    Returns
    -------
    rank : int
        The rank.
    VH : array
        The (adjoint) right singular vectors if ``get_vectors=True``.
    """
    if k_max is None:
        k_max = min(A.shape)
    if eps <= 0.0:
        return k_max

    use_sli = (use_sli and (k_max == min(A.shape)) and
               isdouble(A.dtype) and not get_vectors)
    if use_sli:
        return sla.interpolative.estimate_rank(A, eps)

    if A.shape[0] < A.shape[1]:
        A = A.T
        if get_vectors:
            raise ValueError
    if AH is None:
        AH = dag(A)

    _, s, VH = rsvd_iterate(A, eps, q=q, p=p, G0=G0, AH=AH, use_qb=use_qb,
                            k_start=k_start, k_max=k_max, k_incr=k_incr)

    rank = count_svdvals_needed(s, eps)

    if get_vectors:
        return rank, VH[:rank, :]
    return rank


def maybe_flip(UsV, flipped):
    # if only singular values or only tranposing do nothing
    if not (isinstance(UsV, tuple) and flipped):
        return UsV
    U, s, V = UsV
    return V.T, s, U.T


def rsvd(A, eps_or_k, compute_uv=True, mode='adapt+block', use_qb=20,
         q=2, p=0, k_max=None, k_start=2, k_incr=1.4, G0=None, AH=None):
    """Fast, randomized, iterative SVD. Adaptive variant of method due
    originally to Halko. This scales as ``log(k)`` rather than ``k`` so can be
    more efficient.

    Parameters
    ----------
    A : operator, shape (m, n)
        The operator to decompose.
    eps_or_k : float or int
        Either the relative precision or the number of singular values to
        target. If precision, this is relative to the largest singular value.
    compute_uv : bool, optional
        Whether to return the left and right singular vectors.
    mode : {'adapt+block', 'adapt', 'block'}, optional
        How to perform the randomized SVD. If ``eps_or_k`` is an integer then
        this is implicitly 'block' and ignored. Else:

            - 'adapt+block', perform an initial low quality pass to estimate
              the rank of ``A``, then use the subspace and rank from that to
              perform an accurate fully blocked RSVD.
            - 'adapt', just perform the adaptive randomized SVD.

    q : int, optional
        The number of power iterations, increase for accuracy at the expense
        of runtime.
    p : int, optional
        Oversampling factor. Perform projections with this many extra columns
        and then throw then away.
    k_max : int, optional
        Maximum adaptive rank. Default: ``min(A.shape)``.
    k_start : int, optional
        Initial k when increasing rank adaptively.
    k_incr : float, optional
        Adaptive rank increment factor. Increase the k-step (from k_start) by
        this factor each time. Set to 1 to use a constant step.
    G0 : array_like, shape (n, k), optional
        Initial subspace to start iterating on. If not given a random one will
        be generated.

    Returns
    -------
    U, array, shape (m, k)
        Left singular vectors, if ``compute_uv=True``.
    s, array, shape (k,)
        Singular values.
    V, array, shape (k, n)
        Right singular vectors, if ``compute_uv=True``.
    """

    flipped = A.shape[0] < A.shape[1]
    if flipped:
        A = A.T

    # 'block' mode -> just perform single pass random SVD
    if isinstance(eps_or_k, Integral):
        UsV = rsvd_core(A, eps_or_k, q=q, p=p, state=G0, compute_uv=compute_uv)
        return maybe_flip(UsV, flipped)

    if k_max is None:
        k_max = min(A.shape)
    k_max = min(max(1, k_max), min(A.shape))

    if AH is None:
        AH = dag(A)

    adaptive_opts = {'k_start': k_start, 'k_max': k_max, 'k_incr': k_incr,
                     'use_qb': use_qb, 'AH': AH, 'G0': G0}

    # 'adapt' mode -> rank adaptively perform SVD to low accuracy
    if mode == 'adapt':
        UsV = rsvd_iterate(A, eps_or_k, q=q, p=p,
                           compute_uv=compute_uv, **adaptive_opts)

    # 'adapt+block' mode -> use first pass to find rank, then use blocking mode
    elif mode == 'adapt+block':

        # estimate both rank and get approximate spanning vectors
        k, VH = estimate_rank(A, eps_or_k, get_vectors=True, **adaptive_opts)

        # reuse vectors to effectively boost number of power iterations by one
        UsV = rsvd_core(A, k, q=max(q - 1, 0), p=p, AH=AH,
                        state=dag(VH), compute_uv=compute_uv)

    else:
        raise ValueError("``mode`` must  be one of {'adapt+block', 'adapt'} or"
                         " ``k`` should be a integer to use 'block' mode.")

    return maybe_flip(UsV, flipped)



import time
import functools

import tqdm
import numpy as np

from .tensor_core import TensorNetwork


_TORCH_DEVICE = None


def _get_torch_and_device(dev=None):
    global _TORCH_DEVICE

    if dev is not None:
        assert dev in ['cpu', 'cuda']
        import torch
        _TORCH_DEVICE = torch, dev

    elif _TORCH_DEVICE is None:
        import torch
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        _TORCH_DEVICE = torch, device

    return _TORCH_DEVICE


def variable(x):
    torch, device = _get_torch_and_device()
    return torch.tensor(x, requires_grad=True, device=device)


def constant(x):
    torch, device = _get_torch_and_device()
    return torch.tensor(x, requires_grad=False, device=device)


def variable_tn(tn):
    var_tn = tn.copy()
    var_tn.apply_to_arrays(variable)
    return var_tn


def constant_tn(tn):
    const_tn = tn.copy()
    const_tn.apply_to_arrays(constant)
    return const_tn


def parse_network_to_torch(tn, constant_tags):

    tn_torch = tn.copy()
    variables = []
    iscomplex = False

    for t in tn_torch:

        # check if tensor has any of the constant tags
        if t.tags & constant_tags:
            t.modify(data=constant(t.data))

        # treat re and im parts as separate variables
        elif issubclass(t.dtype.type, np.complexfloating):
            iscomplex = True
            raise TypeError("Torch does not yet support complex data types.")

        else:
            torch_data = variable(t.data)
            variables.append(torch_data)
            t.modify(data=torch_data)

    return tn_torch, variables, iscomplex

def abs_max_grad(tn):
    """Inspect the gradient of every tensor element in the TensorNetwork
    and return the gradient with the largest magnitude (useful for checking
    convergence)."""

    torch, device = _get_torch_and_device()
    views = []
    for t in tn:
        p = t.data
        if p.grad is None:
            view = p.new(p.numel()).zero_()
        elif p.grad.is_sparse:
            view = p.grad.to_dense().view(-1)
        else:
            view = p.grad.view(-1)
        views.append(view)
    flat_grad = torch.cat(views, 0)
    return flat_grad.abs().max()


class TNOptimizer:
    """Optimize a tensor network's arrays using pytorch.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to optimize, every tensor in this will be treated
        as a variable.
    loss_fn : callable
        The target loss function to minimize. Should accept ``tn`` (backed by
        pytorch tensors) or the output of ``norm_fn``.
    norm_fn : callable, optional
        A function to normalize ``tn`` before being passed to ``loss_fn``. 
        This offers very
        general support, so it can be used
        to enforce any constraints such as normalization or unitarity. The
        penalty of using this level of generality is that all operations
        that occur within ``norm_fn`` will be added to the computation graph,
        and thus differentiated during backpropagation. Note that due to the
        generality of this function, it does not actually keep the parameters
        under optimization within the normalized manifold, but instead just
        keeps the evaluation of loss_fn within the normalized manifold. This
        can lead to numerical instability in large TNs where the norm may
        explode after many iterations.
    norm_fn_scalar: callable, optional
        A function that accepts ``tn`` and computes the scalar ``x`` by which
        to multiply each tensor in ``tn`` in order to normalize it. If ``tn``
        is normalized by simple scalar multiplication of the tensors, it is
        recommended to use this function instead of ``norm_fn``. In this case,
        the computation of ``x`` and the rescaling of the tensor values can be 
        performed outside the scope of the
        differentiable computation graph. This has two important advantages:
        1) It can lead to reductions in cpu time and memory usage when compared
        to ``norm_fn``. 2) It can actually modify the parameters under optimization
        to keep them normalized throughout the procedure, increasing numerical
        stability. If desired, both ``norm_fn`` and ``norm_fn_scalar`` can
        be used at the same time to enforce general constraints on ``tn`` and
        then normalize it with scalar multiplication. In this case, note that
        ``norm_fn_scalar`` is applied first, and then ``norm_fn``.
    loss_constants : dict_like, optional
        Extra constant arguments to supply to ``loss_fn`` and be converted to
        tensorflow constant tensors. Can be individual arrays or tensor
        networks.
    loss_kwargs : dict_like, optional
        Other kwargs to supply to ``loss_fn`` that are not arrays or tensor
        networks.
    constant_tags : sequence of str, optional
        Treat any tensors *within* ``tn`` with these tags as constant.
    optimizer : str, optional
        Which optimizer to use, default: ``'Adam'``.
        This should be an optimizer that can be found in the
        ``torch.optim`` submodule.
    learning_rate : float, optional
        The learning rate to apply to use the optimizer with. You can
        dynamically change this between ``optimize`` calls.
    loss_target : float, optional
        If supplied, stop optimizing once this loss is reached.
    tol_change : float, optional
        Convergence threshold on the function value. If the value of the loss 
        function changes by less than ``tol_change`` between subsequent optimization 
        steps, the optimization will terminate.
    tol_grad : float, optional
        Convergence threshold on the gradients. If the absolute value of the 
        gradient with the largest magnitude is smaller than ``tol_grad``, the
        optimization will terminate.
    device : string, optional
        Whether to use ``'cpu'`` or ``'cuda'``, default: ``None``.
        If ``None``, then gpu (``'cuda'``) will be used if it can be
        detected, otherwise ``'cpu'`` will be used.
    progbar : bool, optional
        "True" show live progress of the optimization using a progress bar. 
        "False" runs the optimization silently. For simple line-by-line
        printing of the progress, you can set this option to the string "simple".


    General usage examples can be seen in the documentation
    for ``tensor.optimize_tensorflow.TNOptimizer``.
    """


    def __init__(
        self,
        tn,
        loss_fn,
        norm_fn=None,
        norm_fn_scalar=None,
        loss_constants=None,
        loss_kwargs=None,
        constant_tags=None,
        optimizer='Adam',
        learning_rate=0.01,
        loss_target=None,
        tol_change=1e-9,
        tol_grad=1e-6,
        device=None,
        progbar=True
    ):
        torch, _ = _get_torch_and_device(device)

        self.tn = tn
        self.optimizer = optimizer
        self._learning_rate = learning_rate
        self.loss_target = loss_target
        self.tol_change = tol_change
        self.tol_grad = tol_grad
        assert progbar in [True, False, 'simple']
        self.progbar = progbar
        self.constant_tags = (set() if constant_tags is None
                              else set(constant_tags))
        
        if norm_fn is None:
            def norm_fn(x):
                return x

        self.norm_fn = norm_fn
        self.norm_fn_scalar = norm_fn_scalar

        self.loss_constants = {}
        if loss_constants is not None:
            for k, v in loss_constants.items():
                # check if tensor network supplied
                if isinstance(v, TensorNetwork):
                    # convert it to constant pytorch TN
                    self.loss_constants[k] = constant_tn(v)
                else:
                    self.loss_constants[k] = constant(v)
        self.loss_kwargs = {} if loss_kwargs is None else dict(loss_kwargs)
        self.loss_fn = functools.partial(
            loss_fn, **self.loss_constants, **self.loss_kwargs
        )
        self.loss = None
        self._n = 0

        # make pytorch version of network and gather variables etc.
        self.tn_opt, self.variables, self.iscomplex = parse_network_to_torch(
            tn, self.constant_tags
        )

        self.optimizer = getattr(torch.optim, optimizer)(self.variables,
                                                         lr=learning_rate)

    def closure(self):
        self.optimizer.zero_grad()
        tn = self.norm_fn(self.tn_opt) # MJO: can't move this out of closure
                                       # and down into optimizer() along with 
                                       # norm_fn_scalar because we need
                                       # norm_fn() to be a STRICTLY in-place
                                       # modification of tn_opt for that 
                                       # to work. I'm not sure how we can
                                       # guarantee that for a norm_fn()
                                       # with completely general functionality
        self.loss = self.loss_fn(tn)
        self.loss.backward()
        return self.loss

    @property
    def learning_rate(self):
        return self._learning_rate

    @learning_rate.setter
    def learning_rate(self, lr):
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = lr

    @property
    def nevals(self):
        """The number of gradient evaluations.
        """
        return self._n

    def _maybe_start_timer(self, max_time):
        if max_time is None:
            return
        else:
            self._time_start = time.time()

    def _time_should_stop(self, max_time):
        if max_time is None:
            return False
        else:
            return (time.time() - self._time_start) > max_time

    def _get_tn_opt_numpy(self):
        tn_opt_numpy = self.norm_fn(self.tn_opt)
        tn_opt_numpy.apply_to_arrays(lambda x: x.cpu().detach().numpy())
        return tn_opt_numpy

    def optimize(self, max_steps, max_time=None):

        # perform the optimization with live progress
        if self.progbar is True:
            pbar = tqdm.tqdm(total=max_steps)
        else:
            pbar = tqdm.tqdm(total=max_steps, disable=True)
        self._maybe_start_timer(max_time)

        try:
            for it in range(max_steps):
                if self.progbar == 'simple':
                    t0 = time.time()

                # keep the tn actually normalized during the optimization
                # by doing STRICTLY IN-PLACE modification of tn_opt
                if self.norm_fn_scalar is not None:
                    torch, _ = _TORCH_DEVICE
                    with torch.no_grad():
                        fac = self.norm_fn_scalar(self.tn_opt)
                        for tensor in self.tn_opt:
                            tensor._data *= fac

                self.optimizer.step(self.closure)
                pbar.set_description(f"{self.loss}")
                pbar.update()
                self._n += 1

                if self.progbar == 'simple':
                    t1 = time.time()
                    print('loss = {}, iter time = {}'.format(self.loss, t1-t0))

                # stopping conditions
                if (self.loss_target is not None):
                    if self.loss < self.loss_target:
                        break
                if self._time_should_stop(max_time):
                    break
                if it > 0:
                    if abs(self.loss - prev_loss) <= self.tol_change:
                        break
                if abs_max_grad(self.tn_opt) <= self.tol_grad:
                    break

                pbar.set_description(f"{self.loss}")
                prev_loss = self.loss

        except Exception as excpt:
            if not isinstance(excpt, KeyboardInterrupt):
                print(excpt)
                raise excpt
        finally:
            pbar.close()

        return self._get_tn_opt_numpy()



import numpy as np
import scipy.linalg as scla
import scipy.linalg.interpolative as sli
from autoray import do, reshape

from ..core import njit
from ..linalg.base_linalg import svds, eigh
from ..linalg.rand_linalg import rsvd, estimate_rank


@njit(['i4(f4[:], f4, i4)', 'i4(f8[:], f8, i4)'])  # pragma: no cover
def _trim_singular_vals(s, cutoff, cutoff_mode):
    """Find the number of singular values to keep of ``s`` given ``cutoff`` and
    ``cutoff_mode``.

    Parameters
    ----------
    s : array
        Singular values.
    cutoff : float
        Cutoff.
    cutoff_mode : {1, 2, 3}
        How to perfrm the trim:

            - 1: ['abs'], trim values below ``cutoff``
            - 2: ['rel'], trim values below ``s[0] * cutoff``
            - 3: ['sum2'], trim s.t. ``sum(s_trim**2) < cutoff``.
            - 4: ['rsum2'], trim s.t. ``sum(s_trim**2) < sum(s**2) * cutoff``.
    """
    if cutoff_mode == 1:
        n_chi = np.sum(s > cutoff)

    elif cutoff_mode == 2:
        n_chi = np.sum(s > cutoff * s[0])

    elif (cutoff_mode == 3) or (cutoff_mode == 4):

        target = cutoff
        if cutoff_mode == 4:
            target *= np.sum(s**2)

        n_chi = s.size
        s2s = 0.0
        for i in range(s.size - 1, -1, -1):
            s2 = s[i]**2
            if not np.isnan(s2):
                s2s += s2
            if s2s > target:
                break
            n_chi -= 1
    else:
        raise ValueError("``cutoff_mode`` not one of {1, 2, 3, 4}.")

    return max(n_chi, 1)


@njit(['f4(f4[:], i4)', 'f8(f8[:], i4)'])  # pragma: no cover
def _renorm_singular_vals(s, n_chi):
    """Find the normalization constant for ``s`` such that the new sum squared
    of the ``n_chi`` largest values equals the sum squared of all the old ones.
    """
    s_tot_keep = 0.0
    s_tot_lose = 0.0
    for i in range(s.size):
        s2 = s[i]**2
        if not np.isnan(s2):
            if i < n_chi:
                s_tot_keep += s2
            else:
                s_tot_lose += s2
    return ((s_tot_keep + s_tot_lose) / s_tot_keep)**0.5


@njit  # pragma: no cover
def _trim_and_renorm_SVD(U, s, V, cutoff, cutoff_mode, max_bond, absorb):
    if cutoff > 0.0:
        n_chi = _trim_singular_vals(s, cutoff, cutoff_mode)

        if max_bond > 0:
            n_chi = min(n_chi, max_bond)

        if n_chi < s.size:
            norm = _renorm_singular_vals(s, n_chi)
            s = s[:n_chi] * norm
            U = U[..., :n_chi]
            V = V[:n_chi, ...]

    s = np.ascontiguousarray(s)

    if absorb == -1:
        U *= s.reshape((1, -1))
    elif absorb == 1:
        V *= s.reshape((-1, 1))
    else:
        s **= 0.5
        U *= s.reshape((1, -1))
        V *= s.reshape((-1, 1))

    return U, V


@njit  # pragma: no cover
def _svd_nb(x, cutoff=-1.0, cutoff_mode=3, max_bond=-1, absorb=0):
    """SVD-decomposition.
    """
    U, s, V = np.linalg.svd(x, full_matrices=False)
    return _trim_and_renorm_SVD(U, s, V, cutoff, cutoff_mode, max_bond, absorb)


def _svd_alt(x, cutoff=-1.0, cutoff_mode=3, max_bond=-1, absorb=0):
    """SVD-decompt using alternate scipy driver.
    """
    U, s, V = scla.svd(x, full_matrices=False, lapack_driver='gesvd')
    return _trim_and_renorm_SVD(U, s, V, cutoff, cutoff_mode, max_bond, absorb)


def _svd_numpy(x, cutoff=-1.0, cutoff_mode=3, max_bond=-1, absorb=0):
    args = (x, cutoff, cutoff_mode, max_bond, absorb)

    try:
        return _svd_nb(*args)

    except (scla.LinAlgError, ValueError) as e:  # pragma: no cover

        if isinstance(e, scla.LinAlgError) or 'converge' in str(e):
            import warnings
            warnings.warn("TN SVD failed, trying again with alternate driver.")

            return _svd_alt(*args)

        raise e


def _svd(x, cutoff=-1.0, cutoff_mode=3, max_bond=-1, absorb=0):
    if isinstance(x, np.ndarray):
        return _svd_numpy(x, cutoff, cutoff_mode, max_bond, absorb)

    U, s, VH = do('linalg.svd', x)

    if cutoff > 0.0:
        if cutoff_mode == 1:
            n_chi = do('count_nonzero', s > cutoff)

        elif cutoff_mode == 2:
            n_chi = do('count_nonzero', s > cutoff * s[0])

        elif cutoff_mode in (3, 4):
            s2 = s * s
            cs2 = do('cumsum', s2, 0)
            tot = cs2[-1]

            if cutoff_mode == 3:
                n_chi = do('count_nonzero', (tot - cs2) > cutoff) + 1
            else:
                n_chi = do('count_nonzero', cs2 < (1 - cutoff) * tot) + 1

        n_chi = max(n_chi, 1)
        if max_bond > 0:
            n_chi = min(n_chi, max_bond)

        if n_chi < s.shape[0]:
            s = s[:n_chi]
            U = U[..., :n_chi]
            VH = VH[:n_chi, ...]

        if cutoff_mode in (3, 4):
            norm = (tot / cs2[n_chi - 1]) ** 0.5
            s *= norm

    elif max_bond > 0:
        s = s[:max_bond]
        U = U[..., :max_bond]
        VH = VH[:max_bond, ...]

    if absorb == -1:
        U = U * reshape(s, (1, -1))
    elif absorb == 1:
        VH = VH * reshape(s, (-1, 1))
    else:
        s **= 0.5
        U = U * reshape(s, (1, -1))
        VH = VH * reshape(s, (-1, 1))

    return U, VH


def _svdvals(x):
    """SVD-decomposition, but return singular values only.
    """
    return np.linalg.svd(x, full_matrices=False, compute_uv=False)


@njit  # pragma: no cover
def dag(x):
    """Hermitian conjugate.
    """
    return np.conjugate(x).T


@njit  # pragma: no cover
def _eig(x, cutoff=-1.0, cutoff_mode=3, max_bond=-1, absorb=0):
    """SVD-split via eigen-decomposition.
    """
    if x.shape[0] > x.shape[1]:
        # Get sU, V
        s2, V = np.linalg.eigh(dag(x) @ x)
        U = x @ V
        V = dag(V)

        # Check if want U, sV
        if absorb == 1:
            s = s2**0.5
            U /= s.reshape((1, -1))
            V *= s.reshape((-1, 1))
        # Or sqrt(s)U, sqrt(s)V
        elif absorb == 0:
            sqrts = s2**0.25
            U /= sqrts.reshape((1, -1))
            V *= sqrts.reshape((-1, 1))

    else:
        # Get U, sV
        s2, U = np.linalg.eigh(x @ dag(x))
        V = dag(U) @ x

        # Check if want sU, V
        if absorb == -1:
            s = s2**0.5
            U *= s.reshape((1, -1))
            V /= s.reshape((-1, 1))

        # Or sqrt(s)U, sqrt(s)V
        elif absorb == 0:
            sqrts = s2**0.25
            U *= sqrts.reshape((1, -1))
            V /= sqrts.reshape((-1, 1))

    # eigh produces ascending eigenvalue order -> slice opposite to svd
    if cutoff > 0.0:
        s = s2[::-1]**0.5
        n_chi = _trim_singular_vals(s, cutoff, cutoff_mode)

        if max_bond > 0:
            n_chi = min(n_chi, max_bond)

        if n_chi < s.size:
            norm = _renorm_singular_vals(s, n_chi)
            U = U[..., -n_chi:]
            V = V[-n_chi:, ...]

            if absorb == -1:
                U *= norm
            elif absorb == 0:
                U *= norm**0.5
                V *= norm**0.5
            else:
                V *= norm

    return U, V


@njit
def _svdvals_eig(x):  # pragma: no cover
    """SVD-decomposition via eigen, but return singular values only.
    """
    if x.shape[0] > x.shape[1]:
        s2 = np.linalg.eigvalsh(dag(x) @ x)
    else:
        s2 = np.linalg.eigvalsh(x @ dag(x))
    return s2**0.5


@njit  # pragma: no cover
def _eigh(x, cutoff=-1.0, cutoff_mode=3, max_bond=-1, absorb=0):
    """SVD-decomposition, using hermitian eigen-decomposition, only works if
    ``x`` is hermitian.
    """
    s, U = np.linalg.eigh(x)
    s, U = s[::-1], U[:, ::-1]  # make sure largest singular value first

    V = np.sign(s).reshape(-1, 1) * dag(U)
    s = np.abs(s)
    return _trim_and_renorm_SVD(U, s, V, cutoff, cutoff_mode, max_bond, absorb)


@njit  # pragma: no cover
def _numba_cholesky(x, cutoff=-1, cutoff_mode=3, max_bond=-1, absorb=0):
    """SVD-decomposition, using cholesky decomposition, only works if
    ``x`` is positive definite.
    """
    L = np.linalg.cholesky(x)
    return L, dag(L)


def _cholesky(x, cutoff=-1, cutoff_mode=3, max_bond=-1, absorb=0):
    try:
        return _numba_cholesky(x, cutoff, cutoff_mode, max_bond, absorb)
    except np.linalg.LinAlgError as e:
        if cutoff < 0:
            raise e
        # try adding cutoff identity - assuming it is approx allowable error
        xi = x + 2 * cutoff * np.eye(x.shape[0])
        return _numba_cholesky(xi, cutoff, cutoff_mode, max_bond, absorb)


def _choose_k(x, cutoff, max_bond):
    """Choose the number of singular values to target.
    """
    d = min(x.shape)

    if cutoff != 0.0:
        k = estimate_rank(x, cutoff, k_max=None if max_bond < 0 else max_bond)
    else:
        k = min(d, max_bond)

    # if computing more than half of spectrum then just use dense method
    return 'full' if k > d // 2 else k


def _svds(x, cutoff=0.0, cutoff_mode=2, max_bond=-1, absorb=0):
    """SVD-decomposition using iterative methods. Allows the
    computation of only a certain number of singular values, e.g. max_bond,
    from the get-go, and is thus more efficient. Can also supply
    ``scipy.sparse.linalg.LinearOperator``.
    """
    k = _choose_k(x, cutoff, max_bond)

    if k == 'full':
        if not isinstance(x, np.ndarray):
            x = x.to_dense()
        return _svd(x, cutoff, cutoff_mode, max_bond, absorb)

    U, s, V = svds(x, k=k)
    return _trim_and_renorm_SVD(U, s, V, cutoff, cutoff_mode, max_bond, absorb)


def _isvd(x, cutoff=0.0, cutoff_mode=2, max_bond=-1, absorb=0):
    """SVD-decomposition using interpolative matrix random methods. Allows the
    computation of only a certain number of singular values, e.g. max_bond,
    from the get-go, and is thus more efficient. Can also supply
    ``scipy.sparse.linalg.LinearOperator``.
    """
    k = _choose_k(x, cutoff, max_bond)

    if k == 'full':
        if not isinstance(x, np.ndarray):
            x = x.to_dense()
        return _svd(x, cutoff, cutoff_mode, max_bond, absorb)

    U, s, V = sli.svd(x, k)
    V = dag(V)
    return _trim_and_renorm_SVD(U, s, V, cutoff, cutoff_mode, max_bond, absorb)


def _rsvd(x, cutoff=0.0, cutoff_mode=2, max_bond=-1, absorb=0):
    """SVD-decomposition using randomized methods (due to Halko). Allows the
    computation of only a certain number of singular values, e.g. max_bond,
    from the get-go, and is thus more efficient. Can also supply
    ``scipy.sparse.linalg.LinearOperator``.
    """
    U, s, V = rsvd(x, cutoff)
    return _trim_and_renorm_SVD(U, s, V, cutoff, cutoff_mode, max_bond, absorb)


def _eigsh(x, cutoff=0.0, cutoff_mode=2, max_bond=-1, absorb=0):
    """SVD-decomposition using iterative hermitian eigen decomp, thus assuming
    that ``x`` is hermitian. Allows the computation of only a certain number of
    singular values, e.g. max_bond, from the get-go, and is thus more
    efficient. Can also supply ``scipy.sparse.linalg.LinearOperator``.
    """
    k = _choose_k(x, cutoff, max_bond)

    if k == 'full':
        if not isinstance(x, np.ndarray):
            x = x.to_dense()
        return _eigh(x, cutoff, cutoff_mode, max_bond, absorb)

    s, U = eigh(x, k=k)
    s, U = s[::-1], U[:, ::-1]  # make sure largest singular value first
    V = np.sign(s).reshape(-1, 1) * dag(U)
    s = np.abs(s)
    return _trim_and_renorm_SVD(U, s, V, cutoff, cutoff_mode, max_bond, absorb)


@njit  # pragma: no cover
def _qr_numba(x):
    """QR-decomposition.
    """
    Q, R = np.linalg.qr(x)
    return Q, R


def _qr(x):
    if isinstance(x, np.ndarray):
        return _qr_numba(x)
    return do('linalg.qr', x)


@njit  # pragma: no cover
def _lq_numba(x):
    """LQ-decomposition.
    """
    Q, L = np.linalg.qr(x.T)
    return L.T, Q.T


def _lq(x):
    if isinstance(x, np.ndarray):
        return _lq_numba(x)
    Q, L = do('linalg.qr', do('transpose', x))
    return do('transpose', L), do('transpose', Q)



"""DMRG-like variational algorithms, but in tensor network language.
"""

import itertools
import numpy as np

from ..utils import progbar
from ..core import prod
from ..linalg.base_linalg import eigh, IdentityLinearOperator
from .tensor_core import (
    Tensor,
    TensorNetwork,
    tensor_contract,
    TNLinearOperator,
    asarray,
)


def get_default_opts(cyclic=False):
    """Get the default advanced settings for DMRG.

    Returns
    -------
    default_sweep_sequence : str
        How to sweep. Will be repeated, e.g. "RRL" -> RRLRRLRRL..., default: R.
    bond_compress_method : {'svd', 'eig', ...}
        Method used to compress sites after update.
    bond_compress_cutoff_mode : {'sum2', 'abs', 'rel'}
        How to perform compression truncation.
    bond_expand_rand_strength : float
        In DMRG1, strength of randomness to expand bonds with. Needed to avoid
        singular matrices after expansion.
    local_eig_tol : float
        Relative tolerance to solve inner eigenproblem to, larger = quicker but
        more unstable, default: 1e-3. Note this can be much looser than the
        overall tolerance, the starting point for each local solve is the
        previous state, and the overall accuracy comes from multiple sweeps.
    local_eig_ncv : int
        Number of inner eigenproblem lanczos vectors. Smaller can mean quicker.
    local_eig_backend : {None, 'AUTO', 'SCIPY', 'SLEPC'}
        Which to backend to use for the inner eigenproblem. None or 'AUTO' to
        choose best. Generally ``'SLEPC'`` best if available for large
        problems, but it can't currently handle ``LinearOperator`` Neff as well
        as ``'lobpcg'``.
    local_eig_maxiter : int
        Maximum number of inner eigenproblem iterations.
    local_eig_ham_dense : bool
        Force dense representation of the effective hamiltonian.
    local_eig_EPSType : {'krylovschur', 'gd', 'jd', ...}
        Eigensovler tpye if ``local_eig_backend='slepc'``.
    local_eig_norm_dense : bool
        Force dense representation of the effective norm.
    periodic_segment_size : float or int
        How large (as a proportion if float) to make the 'segments' in periodic
        DMRG. During a sweep everything outside this (the 'long way round') is
        compressed so the effective energy and norm can be efficiently formed.
        Tradeoff: longer segments means having to compress less, but also
        having a shorter 'long way round', meaning that it needs a larger bond
        to represent it and can be 'pseudo-orthogonalized' less effectively.
        0.5 is the largest fraction that makes sense. Set to >= 1.0 to not
        use segmentation at all, which is better for small systems.
    periodic_compress_method : {'isvd', 'svds'}
        Which method to perform the transfer matrix compression with.
    periodic_compress_norm_eps : float
        Precision to compress the norm transfer matrix in periodic systems.
    periodic_compress_ham_eps : float
        Precision to compress the energy transfer matrix in periodic systems.
    periodic_compress_max_bond : int
        The maximum bond to use when compressing transfer matrices.
    periodic_nullspace_fudge_factor : float
        Factor to add to ``Heff`` and ``Neff`` to remove nullspace.
    periodic_canonize_inv_tol : float
        When psuedo-orthogonalizing, an inverse gauge is generated that can be
        very ill-conditioned. This factor controls cutting off the small
        singular values of the gauge to stop this.
    periodic_orthog_tol : float
        When psuedo-orthogonalizing, if the local norm is within this
        distance to 1 (pseudo-orthogonoalized), then the generalized eigen
        decomposition is *not* used, which is much more efficient. If set too
        large the total normalization can become unstable.
    """
    return {
        'default_sweep_sequence': 'R',
        'bond_compress_method': 'svd',
        'bond_compress_cutoff_mode': 'rel' if cyclic else 'sum2',
        'bond_expand_rand_strength': 1e-6,
        'local_eig_tol': 1e-3,
        'local_eig_ncv': 4,
        'local_eig_backend': None,
        'local_eig_maxiter': None,
        'local_eig_EPSType': None,
        'local_eig_ham_dense': None,
        'local_eig_norm_dense': None,
        'periodic_segment_size': 1 / 2,
        'periodic_compress_method': 'isvd',
        'periodic_compress_norm_eps': 1e-6,
        'periodic_compress_ham_eps': 1e-6,
        'periodic_compress_max_bond': -1,
        'periodic_nullspace_fudge_factor': 1e-12,
        'periodic_canonize_inv_tol': 1e-10,
        'periodic_orthog_tol': 1e-6,
    }


class MovingEnvironment:
    r"""Helper class for efficiently moving the effective 'environment' of a
    few sites in a 1D tensor network. E.g. for ``begin='left', bsz=2``, this
    initializes the right environments like so::

        n - 1: o-o-o-     -o-o-o
               | | |       | | |
               H-H-H- ... -H-H-H
               | | |       | | |
               o-o-o-     -o-o-o

        n - 2: o-o-o-     -o-o\
               | | |       | | o
               H-H-H- ... -H-H-H
               | | |       | | o
               o-o-o-     -o-o/

        n - 3: o-o-o-     -o\
               | | |       | oo
               H-H-H- ... -H-HH
               | | |       | oo
               o-o-o-     -o/

        ...

        0    : o-o\
               | | oo   ooo
               H-H-HH...HHH
               | | oo   ooo
               o-o/

    which can then be used to efficiently generate the left environments as
    each site is updated. For example if ``bsz=2`` and the environements have
    been shifted many sites into the middle, then ``MovingEnvironment()``
    returns something like::

             <---> bsz sites
             /o-o\
        ooooo | | ooooooo
        HHHHH-H-H-HHHHHHH
        ooooo | | ooooooo
             \o-o/
        0 ... i i+1 ... n-1

    For periodic systems ``MovingEnvironment`` approximates the 'long
    way round' transfer matrices. E.g consider replacing segment B
    (to arbitrary precision) with an SVD::

        /-----------------------------------------------\
        +-A-A-A-A-A-A-A-A-A-A-A-A-B-B-B-B-B-B-B-B-B-B-B-+
          | | | | | | | | | | | | | | | | | | | | | | |           ==>
        +-A-A-A-A-A-A-A-A-A-A-A-A-B-B-B-B-B-B-B-B-B-B-B-+
        \-----------------------------------------------/

        +~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
        |   /-A-A-A-A-A-A-A-A-A-A-A-A-\   |                       ==>
        +~<BL | | | | | | | | | | | | BR>~+
            \-A-A-A-A-A-A-A-A-A-A-A-A-/
              ^                     ^
        segment_start          segment_stop - 1

        +~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
        |   /-A-A-\                        |                      ==>
        +~<BL | |  AAAAAAAAAAAAAAAAAAAABR>~+
            \-A-A-/
              ...
            <-bsz->

        +~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
        |               /-A-A-\           |                       ==>
        +~<BLAAAAAAAAAAA  | |  AAAAAAABR>~+
                        \-A-A-/
                          i i+1
             -----sweep--------->

    Can then contract and store left and right environments for efficient
    sweeping just as in non-periodic case. If the segment is long enough (50+)
    sites, often only 1 singular value is needed, and thus the efficiency is
    the same as for OBC.

    Parameters
    ----------
    tn : TensorNetwork
        The 1D tensor network, should be closed, i.e. an overlap of some sort.
    begin : {'left', 'right'}
        Which side to start at and sweep from.
    bsz : int
        The number of sites that form the the 'non-environment',
        e.g. 2 for DMRG2.
    ssz : float or int, optional
        The size of the segment to use, if float, the proportion. Default: 1/2.
    eps : float, optional
        The tolerance to approximate the transfer matrix with. See
        :meth:`~quimb.tensor.TensorNetwork.replace_with_svd`.
    cyclic : bool, optional
        Whether this is a periodic ``MovingEnvironment``.
    segment_callbacks : sequence of callable, optional
        Functions with signature ``callback(start, stop, self.begin)``, to be
        called every time a new segment is initialized.
    method : {'isvd', 'svds', ...}, optional
        How to performt the transfer matrix compression. See
        :meth:`~quimb.tensor.TensorNetwork.replace_with_svd`.
    max_bond : , optional
        If > 0, the maximum bond of the compressed transfer matrix.
    norm : bool, optional
        If True, treat this ``MovingEnvironment`` as the state overlap, which
        enables a few extra checks.

    Notes
    -----
    Does not necessarily need to be an operator overlap tensor network. Useful
    for any kind of sweep where only local tensor updates are being made. Note
    that *only* the current site is completely up-to-date and can be modified
    with changes meant to propagate.
    """

    def __init__(self, tn, begin, bsz, *, cyclic=False, segment_callbacks=None,
                 ssz=0.5, eps=1e-8, method='isvd', max_bond=-1, norm=False):

        self.tn = tn.copy(virtual=True)
        self.begin = begin
        self.bsz = bsz
        self.cyclic = cyclic

        if callable(segment_callbacks):
            self.segment_callbacks = (segment_callbacks,)
        else:
            self.segment_callbacks = segment_callbacks

        self.n = tn.nsites
        self.structure = tn.structure

        if self.cyclic:
            self.eps = eps
            self.method = method
            self.max_bond = max_bond
            self.norm = norm
            self.bond_sizes = []

            if isinstance(ssz, float):
                # this logic essentially makes sure that segments prefer
                #     overshooting e.g ssz=1/3 with n=100 produces segments of
                #     length 34, to avoid a final segement of length 1.
                self._ssz = int(self.n * ssz + self.n % int(1 / ssz))
            else:
                self._ssz = ssz

            self.segmented = self._ssz < self.n
            # will still split system in half but no compression or callbacks
            if not self.segmented:
                self._ssz = int(self.n / 2 + self.n % 2)

            start, stop = {
                'left': (0, self._ssz),
                'right': (self.n - self._ssz, self.n)
            }[begin]
        else:
            self.segmented = False
            start, stop = (0, self.n - self.bsz + 1)

        self.init_segment(begin, start, stop)

    def site_tag(self, i):
        return self.structure.format(i % self.n)

    def init_segment(self, begin, start, stop):
        """Initialize the environments in ``range(start, stop)`` so that one
        can start sweeping from the side defined by ``begin``.
        """
        if (start >= self.n) or (stop < 0):
            start, stop = start % self.n, stop % self.n

        self.segment = range(start, stop)
        self.init_non_segment(start, stop + self.bsz // 2)

        if begin == 'left':

            tags_initital = ['_RIGHT'] + [self.site_tag(stop - 1 + b)
                                          for b in range(self.bsz)]
            self.envs = {stop - 1: self.tnc.select(tags_initital, which='any')}

            for i in reversed(range(start, stop - 1)):
                # add a new site to previous env, and contract one site
                self.envs[i] = self.envs[i + 1].copy(virtual=True)
                self.envs[i] |= self.tnc.select(i)
                self.envs[i] ^= ('_RIGHT', self.site_tag(i + self.bsz))

            self.envs[start] |= self.tnc['_LEFT']
            self.pos = start

        elif begin == 'right':

            tags_initital = ['_LEFT'] + [self.site_tag(start + b)
                                         for b in range(self.bsz)]
            self.envs = {start: self.tnc.select(tags_initital, which='any')}

            for i in range(start + 1, stop):
                # add a new site to previous env, and contract one site
                self.envs[i] = self.envs[i - 1].copy(virtual=True)
                self.envs[i] |= self.tnc.select(i + self.bsz - 1)
                self.envs[i] ^= ('_LEFT', self.site_tag(i - 1))

            self.envs[i] |= self.tnc['_RIGHT']
            self.pos = stop - 1

        else:
            raise ValueError("``begin`` must be 'left' or 'right'.")

    def init_non_segment(self, start, stop):
        """Compress and label the effective env not in ``range(start, stop)``
        if cyclic, else just add some dummy left and right end pieces.
        """
        self.tnc = self.tn.copy(virtual=True)

        if not self.segmented:
            if not self.cyclic:
                # generate dummy left and right envs
                self.tnc |= Tensor(1.0, (), {'_LEFT'}).astype(self.tn.dtype)
                self.tnc |= Tensor(1.0, (), {'_RIGHT'}).astype(self.tn.dtype)
                return

            # if cyclic just contract other section and tag
            self.tnc |= Tensor(1.0, (), {'_LEFT'}).astype(self.tn.dtype)
            self.tnc.contract(slice(stop, start + self.n), inplace=True)
            self.tnc.add_tag('_RIGHT', where=stop + 1)
            return

        # replicate all tags on end pieces apart from site number
        ltags = {'_LEFT', *self.tnc.select(start - 1).tags}
        ltags.remove(self.site_tag(start - 1))
        rtags = {'_RIGHT', *self.tnc.select(stop).tags}
        rtags.remove(self.site_tag(stop))

        # for example, pseudo orthogonalization if cyclic
        if self.segment_callbacks is not None:
            for callback in self.segment_callbacks:
                callback(start, stop, self.begin)

        opts = {
            'keep_tags': False,
            'ltags': ltags,
            'rtags': rtags,
            'eps': self.eps,
            'method': self.method,
            'max_bond': self.max_bond,
            'inplace': True,
        }

        self.tnc.replace_section_with_svd(start, stop, which='!any', **opts)

        self.bond_sizes.append(
            self.tnc['_LEFT'].shared_bond_size(self.tnc['_RIGHT']))

        if self.norm:
            # ensure that expectation still = 1 after approximation
            # section left can still be pretty long so do structured contract
            tnn = self.tnc.copy()
            tnn ^= ['_LEFT', self.site_tag(start)]
            tnn ^= ['_RIGHT', self.site_tag(stop - 1)]
            norm = (tnn ^ slice(start, stop)) ** 0.5

            self.tnc['_LEFT'] /= norm
            self.tnc['_RIGHT'] /= norm

    def move_right(self):
        i = (self.pos + 1) % self.n

        # generate a new segment if we go over the border
        if i not in self.segment:
            if not self.cyclic:
                raise ValueError("For OBC, ``0 <= position <= n - bsz``.")
            self.init_segment('left', i, i + self._ssz)
        else:
            self.pos = i % self.n

        i0 = self.segment.start

        if i >= i0 + 1:
            # insert the updated left env from previous step
            # contract left env with updated site just to left
            new_left = self.envs[i - 1].select(
                ['_LEFT', self.site_tag(i - 1)], which='any')
            self.envs[i] |= new_left ^ all

    def move_left(self):
        i = (self.pos - 1) % self.n

        # generate a new segment if we go over the border
        if i not in self.segment:
            if not self.cyclic:
                raise ValueError("For OBC, ``0 <= position <= n - bsz``.")
            self.init_segment('right', i - self._ssz + 1, i + 1)
        else:
            self.pos = i % self.n

        iN = self.segment.stop

        if i <= iN - 2:
            # insert the updated right env from previous step
            # contract right env with updated site just to right
            new_right = self.envs[i + 1].select(
                ['_RIGHT', self.site_tag(i + self.bsz)], which='any')
            self.envs[i] |= new_right ^ all

    def move_to(self, i):
        """Move this effective environment to site ``i``.
        """

        if self.cyclic:
            # to take account of PBC, rescale so that current pos == n // 2,
            #     then work out if desired i is lower or higher
            ri = (i + (self.n // 2 - self.pos)) % self.n
            direction = 'left' if ri <= self.n // 2 else 'right'
        else:
            direction = 'left' if i < self.pos else 'right'

        while self.pos != i % self.n:
            {'left': self.move_left, 'right': self.move_right}[direction]()

    def __call__(self):
        """Get the current environment.
        """
        return self.envs[self.pos]


def get_cyclic_canonizer(k, b, inv_tol=1e-10):
    """Get a function to use as a callback for ``MovingEnvironment`` that
    approximately orthogonalizes the segments of periodic MPS.
    """
    def cyclic_canonizer(start, stop, begin):
        k.canonize_cyclic(slice(start, stop), bra=b, inv_tol=inv_tol)
        if begin == 'left':
            k.right_canonize(start=stop - 1, stop=start, bra=b)
        else:
            k.left_canonize(start=start, stop=stop - 1, bra=b)

    return cyclic_canonizer


# --------------------------------------------------------------------------- #
#                                  DMRG Base                                  #
# --------------------------------------------------------------------------- #

def parse_2site_inds_dims(k, b, i):
    r"""Sort out the dims and inds of::

        ---O---O---
           |   |

    For use in 2 site algorithms.
    """
    u_bond_ind = k.bond(i, i + 1)
    dims_L, uix_L = zip(*(
        (d, ix)
        for d, ix in zip(k[i].shape, k[i].inds)
        if ix != u_bond_ind
    ))
    dims_R, uix_R = zip(*(
        (d, ix)
        for d, ix in zip(k[i + 1].shape,
                         k[i + 1].inds)
        if ix != u_bond_ind
    ))
    uix = uix_L + uix_R

    l_bond_ind = b.bond(i, i + 1)
    lix_L = tuple(i for i in b[i].inds if i != l_bond_ind)
    lix_R = tuple(i for i in b[i + 1].inds if i != l_bond_ind)
    lix = lix_L + lix_R

    dims = dims_L + dims_R

    return dims, lix_L, lix_R, lix, uix_L, uix_R, uix, l_bond_ind, u_bond_ind


class DMRGError(Exception):
    pass


class DMRG:
    r"""Density Matrix Renormalization Group variational groundstate search.
    Some initialising arguments act as defaults, but can be overidden with
    each solve or sweep. See :func:`~quimb.tensor.tensor_dmrg.get_default_opts`
    for the list of advanced options initialized in the ``opts`` attribute.

    Parameters
    ----------
    ham : MatrixProductOperator
        The hamiltonian in MPO form.
    bond_dims : int or sequence of ints.
        The bond-dimension of the MPS to optimize. If ``bsz > 1``, then this
        corresponds to the maximum bond dimension when splitting the effective
        local groundstate. If a sequence is supplied then successive sweeps
        iterate through, then repeate the final value. E.g.
        ``[16, 32, 64] -> (16, 32, 64, 64, 64, ...)``.
    cutoffs : dict-like
        The cutoff threshold(s) to use when compressing. If a sequence is
        supplied then successive sweeps iterate through, then repeate the final
        value. E.g. ``[1e-5, 1e-7, 1e-9] -> (1e-5, 1e-7, 1e-9, 1e-9, ...)``.
    bsz : {1, 2}
        Number of sites to optimize for locally i.e. DMRG1 or DMRG2.
    which : {'SA', 'LA'}, optional
        Whether to search for smallest or largest real part eigenvectors.
    p0 : MatrixProductState, optional
        If given, use as the initial state.

    Attributes
    ----------
    state : MatrixProductState
        The current, optimized state.
    energy : float
        The current most optimized energy.
    energies : list of float
        The total energy after each sweep.
    local_energies : list of list of float
        The local energies per sweep: ``local_energies[i, j]`` contains the
        local energy found at the jth step of the (i+1)th sweep.
    total_energies : list of list of float
        The total energies per sweep: ``local_energies[i, j]`` contains the
        total energy after the jth step of the (i+1)th sweep.
    opts : dict
        Advanced options e.g. relating to the inner eigensolve or compression,
        see :func:`~quimb.tensor.tensor_dmrg.get_default_opts`.
    (bond_sizes_ham) : list[list[int]]
        If cyclic, the sizes of the energy environement transfer matrix bonds,
        per segment, per sweep.
    (bond_sizes_norm) : list[list[int]]
        If cyclic, the sizes of the norm environement transfer matrix bonds,
        per segment, per sweep.
    """

    def __init__(self, ham, bond_dims, cutoffs=1e-9,
                 bsz=2, which='SA', p0=None):
        self.n = ham.nsites
        self.phys_dim = ham.phys_dim()
        self.bsz = bsz
        self.which = which
        self.cyclic = ham.cyclic
        self._set_bond_dim_seq(bond_dims)
        self._set_cutoff_seq(cutoffs)

        # create internal states and ham
        if p0 is not None:
            self._k = p0.copy()
        else:
            self._k = ham.rand_state(self._bond_dim0)
        self._b = self._k.H
        self.ham = ham.copy()
        self._k.add_tag("_KET")
        self._b.add_tag("_BRA")
        self.ham.add_tag("_HAM")

        # Line up and overlap for energy calc
        self._k.align_(self.ham, self._b)

        # want to contract this multiple times while
        #   manipulating k/b -> make virtual
        self.TN_energy = self._b | self.ham | self._k
        self.energies = []
        self.local_energies = []
        self.total_energies = []

        # if cyclic need to keep track of normalization
        if self.cyclic:
            eye = self.ham.identity()
            eye.add_tag('_EYE')
            self.TN_norm = self._b | eye | self._k

            self.bond_sizes_ham = []
            self.bond_sizes_norm = []

        self.opts = get_default_opts(self.cyclic)

    def _set_bond_dim_seq(self, bond_dims):
        bds = (bond_dims,) if isinstance(bond_dims, int) else tuple(bond_dims)
        self._bond_dim0 = bds[0]
        self._bond_dims = itertools.chain(bds, itertools.repeat(bds[-1]))

    def _set_cutoff_seq(self, cutoffs):
        bds = (cutoffs,) if isinstance(cutoffs, float) else tuple(cutoffs)
        self._cutoffs = itertools.chain(bds, itertools.repeat(bds[-1]))

    @property
    def energy(self):
        return self.energies[-1]

    @property
    def state(self):
        copy = self._k.copy()
        copy.drop_tags('_KET')
        return copy

    # -------------------- standard DMRG update methods --------------------- #

    def _canonize_after_1site_update(self, direction, i):
        """Compress a site having updated it. Also serves to move the
        orthogonality center along.
        """
        if (direction == 'right') and ((i < self.n - 1) or self.cyclic):
            self._k.left_canonize_site(i, bra=self._b)
        elif (direction == 'left') and ((i > 0) or self.cyclic):
            self._k.right_canonize_site(i, bra=self._b)

    def _eigs(self, A, B=None, v0=None):
        """Find single eigenpair, using all the internal settings.
        """
        # intercept generalized eigen
        backend = self.opts['local_eig_backend']
        if (backend is None) and (B is not None):
            backend = 'LOBPCG'

        return eigh(
            A, k=1, B=B, which=self.which, v0=v0,
            backend=backend,
            EPSType=self.opts['local_eig_EPSType'],
            ncv=self.opts['local_eig_ncv'],
            tol=self.opts['local_eig_tol'],
            maxiter=self.opts['local_eig_maxiter'],
            fallback_to_scipy=True)

    def print_energy_info(self, Heff=None, loc_gs=None):
        sweep_num = len(self.energies) + 1
        full_en = self.TN_energy ^ ...
        effv_en = self._eff_ham ^ all

        if Heff is None:
            site_en = "N/A"
        else:
            site_en = (loc_gs.H @ (Heff @ loc_gs)).item()

        print(f"Sweep {sweep_num} -- fullE={full_en} "
              f"effcE={effv_en} siteE={site_en}")

    def print_norm_info(self, i=None):
        sweep_num = len(self.energies) + 1
        full_n = self._k.H @ self._k

        if self.cyclic:
            effv_n = self._eff_norm ^ all
        else:
            effv_n = 'OBC'

        if i is None:
            site_norm = [self._k[i].H @ self._k[i] for i in range(self.n)]
        else:
            site_norm = self._k[i].H @ self._k[i]

        print(f"Sweep {sweep_num} -- fullN={full_n} "
              f"effvN={effv_n} siteN={site_norm}")

    def form_local_ops(self, i, dims, lix, uix):
        """Construct the effective Hamiltonian, and if needed, norm.
        """
        if self.cyclic:
            self._eff_norm = self.ME_eff_norm()
        self._eff_ham = self.ME_eff_ham()

        # choose a rough value at which dense effective ham should not be used
        dense = self.opts['local_eig_ham_dense']
        if dense is None:
            dense = prod(dims) < 800

        dims_inds = {'ldims': dims, 'rdims': dims,
                     'left_inds': lix, 'right_inds': uix}

        # form effective hamiltonian
        if dense:
            # contract remaining hamiltonian and get its dense representation
            Heff = (self._eff_ham ^ '_HAM')['_HAM'].to_dense(lix, uix)
        else:
            Heff = TNLinearOperator(self._eff_ham['_HAM'], **dims_inds)

        # form effective norm
        if self.cyclic:
            fudge = self.opts['periodic_nullspace_fudge_factor']

            neff_dense = self.opts['local_eig_norm_dense']
            if neff_dense is None:
                neff_dense = dense

            # Check if site already pseudo-orthonogal
            site_norm = self._k[i:i + self.bsz].H @ self._k[i:i + self.bsz]
            if abs(site_norm - 1) < self.opts['periodic_orthog_tol']:
                Neff = None

            # else contruct RHS normalization operator
            elif neff_dense:
                Neff = (self._eff_norm ^ '_EYE')['_EYE'].to_dense(lix, uix)
                np.fill_diagonal(Neff, Neff.diagonal() + fudge)
                np.fill_diagonal(Heff, Heff.diagonal() + fudge**0.5)
            else:
                Neff = TNLinearOperator(self._eff_norm['_EYE'], **dims_inds)
                Neff += IdentityLinearOperator(Neff.shape[0], fudge)
                Heff += IdentityLinearOperator(Heff.shape[0], fudge**0.5)

        else:
            Neff = None

        return Heff, Neff

    def post_check(self, i, Neff, loc_gs, loc_en, loc_gs_old):
        """Perform some checks on the output of the local eigensolve.
        """
        if self.cyclic:
            # pseudo-orthogonal
            if Neff is None:
                # just perform leading correction to norm from site_norm
                site_norm = self._k[i:i + self.bsz].H @ self._k[i:i + self.bsz]
                loc_gs *= site_norm ** 0.5
                loc_en *= site_norm
                return loc_en, loc_gs

            loc_en -= self.opts['periodic_nullspace_fudge_factor']**0.5

            # this is helpful for identifying badly behaved numerics
            Neffnorm = (loc_gs.H @ (Neff @ loc_gs)).item()
            if abs(Neffnorm - 1) > 10 * self.opts['local_eig_tol']:
                raise DMRGError(f"Effective norm diverged to {Neffnorm}, "
                                "check that Neff is positive?")

        return loc_en, loc_gs

    def _update_local_state_1site(self, i, direction, **compress_opts):
        r"""Find the single site effective tensor groundstate of::

            >->->->->-/|\-<-<-<-<-<-<-<-<          /|\       <-- uix
            | | | | |  |  | | | | | | | |         / | \
            H-H-H-H-H--H--H-H-H-H-H-H-H-H   =    L--H--R
            | | | | | i|  | | | | | | | |         \i| /
            >->->->->-\|/-<-<-<-<-<-<-<-<          \|/       <-- lix

        And insert it back into the states ``k`` and ``b``, and thus
        ``TN_energy``.
        """
        uix, lix = self._k[i].inds, self._b[i].inds
        dims = self._k[i].shape

        # get local operators
        Heff, Neff = self.form_local_ops(i, dims, lix, uix)

        # get the old local groundstate to use as initial guess
        loc_gs_old = self._k[i].data.ravel()

        # find the local energy and groundstate
        loc_en, loc_gs = self._eigs(Heff, B=Neff, v0=loc_gs_old)

        # perform some minor checks and corrections
        loc_en, loc_gs = self.post_check(i, Neff, loc_gs, loc_en, loc_gs_old)

        # insert back into state and all tensor networks viewing it
        loc_gs = loc_gs.A.reshape(dims)
        self._k[i].modify(data=loc_gs)
        self._b[i].modify(data=loc_gs.conj())

        # normalize - necessary due to loose tolerance eigensolve
        if self.cyclic:
            norm = (self._eff_norm ^ all) ** 0.5
            self._k[i].modify(data=self._k[i].data / norm)
            self._b[i].modify(data=self._b[i].data / norm)

        tot_en = self._eff_ham ^ all

        self._canonize_after_1site_update(direction, i)

        return loc_en.item(), tot_en

    def _update_local_state_2site(self, i, direction, **compress_opts):
        r"""Find the 2-site effective tensor groundstate of::

            >->->->->-/| |\-<-<-<-<-<-<-<-<          /| |\
            | | | | |  | |  | | | | | | | |         / | | \
            H-H-H-H-H--H-H--H-H-H-H-H-H-H-H   =    L--H-H--R
            | | | | |  i i+1| | | | | | | |         \ | | /
            >->->->->-\| |/-<-<-<-<-<-<-<-<          \| |/
                                                 i i+1

        And insert it back into the states ``k`` and ``b``, and thus
        ``TN_energy``.
        """
        dims, lix_L, lix_R, lix, uix_L, uix_R, uix, l_bond_ind, u_bond_ind = \
            parse_2site_inds_dims(self._k, self._b, i)

        # get local operators
        Heff, Neff = self.form_local_ops(i, dims, lix, uix)

        # get the old 2-site local groundstate to use as initial guess
        loc_gs_old = self._k[i].contract(self._k[i + 1]).to_dense(uix)

        # find the 2-site local groundstate and energy
        loc_en, loc_gs = self._eigs(Heff, B=Neff, v0=loc_gs_old)

        # perform some minor checks and corrections
        loc_en, loc_gs = self.post_check(i, Neff, loc_gs, loc_en, loc_gs_old)

        # split the two site local groundstate
        T_AB = Tensor(loc_gs.A.reshape(dims), uix)
        L, R = T_AB.split(left_inds=uix_L, get='arrays', absorb=direction,
                          right_inds=uix_R, **compress_opts)

        # insert back into state and all tensor networks viewing it
        self._k[i].modify(data=L, inds=(*uix_L, u_bond_ind))
        self._b[i].modify(data=L.conj(), inds=(*lix_L, l_bond_ind))
        self._k[i + 1].modify(data=R, inds=(u_bond_ind, *uix_R))
        self._b[i + 1].modify(data=R.conj(), inds=(l_bond_ind, *lix_R))

        # normalize due to compression and insert factor at the correct site
        if self.cyclic:
            #   Right         Left
            #   i  i+1        i  i+1
            # -->~~o--  or  --o~~<--
            #   |  |          |  |
            norm = (self.ME_eff_norm() ^ all) ** 0.5
            next_site = {'right': i + 1, 'left': i}[direction]

            self._k[next_site].modify(data=self._k[next_site].data / norm)
            self._b[next_site].modify(data=self._b[next_site].data / norm)

        tot_en = self._eff_ham ^ all

        return loc_en.item(), tot_en

    def _update_local_state(self, i, **update_opts):
        """Move envs to site ``i`` and dispatch to the correct local updater.
        """
        if self.cyclic:
            # move effective norm first as it can trigger canonize_cyclic etc.
            self.ME_eff_norm.move_to(i)

        self.ME_eff_ham.move_to(i)

        return {
            1: self._update_local_state_1site,
            2: self._update_local_state_2site,
        }[self.bsz](i, **update_opts)

    def sweep(self, direction, canonize=True, verbosity=0, **update_opts):
        r"""Perform a sweep of optimizations, either rightwards::

              optimize -->
                ...
            >->-o-<-<-<-<-<-<-<-<-<-<-<-<-<
            | | | | | | | | | | | | | | | |
            H-H-H-H-H-H-H-H-H-H-H-H-H-H-H-H
            | | | | | | | | | | | | | | | |
            >->-o-<-<-<-<-<-<-<-<-<-<-<-<-<

        or leftwards (`direction='L'`)::

                            <-- optimize
                                      ...
            >->->->->->->->->->->->->-o-<-<
            | | | | | | | | | | | | | | | |
            H-H-H-H-H-H-H-H-H-H-H-H-H-H-H-H
            | | | | | | | | | | | | | | | |
            >->->->->->->->->->->->->-o-<-<

        After the sweep the state is left or right canonized respectively.

        Parameters
        ----------
        direction : {'R', 'L'}
            Sweep from left to right (->) or right to left (<-) respectively.
        canonize : bool, optional
            Canonize the state first, not needed if doing alternate sweeps.
        verbosity : {0, 1, 2}, optional
            Show a progress bar for the sweep.
        update_opts :
            Supplied to ``self._update_local_state``.
        """
        if canonize:
            {'R': self._k.right_canonize,
             'L': self._k.left_canonize}[direction](bra=self._b)

        n, bsz = self.n, self.bsz

        direction, begin, sweep = {
            ('R', False): ('right', 'left', range(0, n - bsz + 1)),
            ('L', False): ('left', 'right', range(n - bsz, -1, -1)),
            ('R', True): ('right', 'left', range(0, n)),
            ('L', True): ('left', 'right', range(n - 1, -1, -1)),
        }[direction, self.cyclic]

        if verbosity:
            sweep = progbar(sweep, ncols=80, total=len(sweep))

        env_opts = {'begin': begin, 'bsz': bsz, 'cyclic': self.cyclic,
                    'ssz': self.opts['periodic_segment_size'],
                    'method': self.opts['periodic_compress_method'],
                    'max_bond': self.opts['periodic_compress_max_bond']}

        if self.cyclic:
            # setup moving norm environment
            nm_opts = {
                **env_opts, 'norm': True,
                'eps': self.opts['periodic_compress_norm_eps'],
                'segment_callbacks': get_cyclic_canonizer(
                    self._k, self._b,
                    inv_tol=self.opts['periodic_canonize_inv_tol']),
            }
            self.ME_eff_norm = MovingEnvironment(self.TN_norm, **nm_opts)

        # setup moving energy environment
        en_opts = {**env_opts, 'eps': self.opts['periodic_compress_ham_eps']}
        self.ME_eff_ham = MovingEnvironment(self.TN_energy, **en_opts)

        # perform the sweep, collecting local and total energies
        local_ens, tot_ens = zip(*[
            self._update_local_state(i, direction=direction, **update_opts)
            for i in sweep
        ])

        if verbosity:
            sweep.close()

        self.local_energies.append(local_ens)
        self.total_energies.append(tot_ens)

        if self.cyclic:
            self.bond_sizes_ham.append(self.ME_eff_ham.bond_sizes)
            self.bond_sizes_norm.append(self.ME_eff_norm.bond_sizes)

        return tot_ens[-1]

    def sweep_right(self, canonize=True, verbosity=0, **update_opts):
        return self.sweep(direction='R', canonize=canonize,
                          verbosity=verbosity, **update_opts)

    def sweep_left(self, canonize=True, verbosity=0, **update_opts):
        return self.sweep(direction='L', canonize=canonize,
                          verbosity=verbosity, **update_opts)

    # ----------------- overloadable 'plugin' style methods ----------------- #

    def _print_pre_sweep(self, i, LR, bd, ctf, verbosity=0):
        """Print this before each sweep.
        """
        if verbosity > 0:
            current_bd = self._k.max_bond()
            msg = "SWEEP-{}, direction={}, max_bond=({}/{}), cutoff:{}"
            print(msg.format(i + 1, LR, current_bd, bd, ctf), flush=True)

    def _compute_post_sweep(self):
        """Compute this after each sweep.
        """
        pass

    def _print_post_sweep(self, converged, verbosity=0):
        """Print this after each sweep.
        """
        if verbosity > 1:
            self._k.show()
        if verbosity > 0:
            msg = "Energy: {} ... {}".format(self.energy, "converged!" if
                                             converged else "not converged.")
            print(msg, flush=True)

    def _check_convergence(self, tol):
        """By default check the absolute change in energy.
        """
        if len(self.energies) < 2:
            return False
        return abs(self.energies[-2] - self.energies[-1]) < tol

    # -------------------------- main solve driver -------------------------- #

    def solve(self,
              tol=1e-4,
              bond_dims=None,
              cutoffs=None,
              sweep_sequence=None,
              max_sweeps=10,
              verbosity=0):
        """Solve the system with a sequence of sweeps, up to a certain
        absolute tolerance in the energy or maximum number of sweeps.

        Parameters
        ----------
        tol : float, optional
            The absolute tolerance to converge energy to.
        bond_dims : int or sequence of int
            Overide the initial/current bond_dim sequence.
        cutoffs : float of sequence of float
            Overide the initial/current cutoff sequence.
        sweep_sequence : str, optional
            String made of 'L' and 'R' defining the sweep sequence, e.g 'RRL'.
            The sequence will be repeated until ``max_sweeps`` is reached.
        max_sweeps : int, optional
            The maximum number of sweeps to perform.
        verbosity : {0, 1, 2}, optional
            How much information to print about progress.

        Returns
        -------
        converged : bool
            Whether the algorithm has converged to ``tol`` yet.
        """
        verbosity = int(verbosity)

        # Possibly overide the default bond dimension, cutoff, LR sequences.
        if bond_dims is not None:
            self._set_bond_dim_seq(bond_dims)
        if cutoffs is not None:
            self._set_cutoff_seq(cutoffs)
        if sweep_sequence is None:
            sweep_sequence = self.opts['default_sweep_sequence']

        RLs = itertools.cycle(sweep_sequence)
        previous_LR = '0'

        for _ in range(max_sweeps):
            # Get the next direction, bond dimension and cutoff
            LR, bd, ctf = next(RLs), next(self._bond_dims), next(self._cutoffs)
            self._print_pre_sweep(len(self.energies), LR,
                                  bd, ctf, verbosity=verbosity)

            # if last sweep was in opposite direction no need to canonize
            canonize = False if LR + previous_LR in {'LR', 'RL'} else True
            # need to manually expand bond dimension for DMRG1
            if self.bsz == 1:
                self._k.expand_bond_dimension(
                    bd, bra=self._b,
                    rand_strength=self.opts['bond_expand_rand_strength'])

            # inject all options and defaults
            sweep_opts = {
                'canonize': canonize,
                'max_bond': bd,
                'cutoff': ctf,
                'cutoff_mode': self.opts['bond_compress_cutoff_mode'],
                'method': self.opts['bond_compress_method'],
                'verbosity': verbosity,
            }

            # perform sweep, any plugin computations
            self.energies.append(self.sweep(direction=LR, **sweep_opts))
            self._compute_post_sweep()

            # check convergence
            converged = self._check_convergence(tol)
            self._print_post_sweep(converged, verbosity=verbosity)
            if converged:
                break

            previous_LR = LR

        return converged


class DMRG1(DMRG):
    """Simple alias of one site ``DMRG``.
    """
    __doc__ += DMRG.__doc__

    def __init__(self, ham, which='SA', bond_dims=None, cutoffs=1e-8, p0=None):

        if bond_dims is None:
            bond_dims = range(10, 1001, 10)

        super().__init__(ham, bond_dims=bond_dims, cutoffs=cutoffs,
                         which=which, p0=p0, bsz=1)


class DMRG2(DMRG):
    """Simple alias of two site ``DMRG``.
    """
    __doc__ += DMRG.__doc__

    def __init__(self, ham, which='SA', bond_dims=None, cutoffs=1e-8, p0=None):

        if bond_dims is None:
            bond_dims = [8, 16, 32, 64, 128, 256, 512, 1024]

        super().__init__(ham, bond_dims=bond_dims, cutoffs=cutoffs,
                         which=which, p0=p0, bsz=2)


# --------------------------------------------------------------------------- #
#                                    DMRGX                                    #
# --------------------------------------------------------------------------- #

class DMRGX(DMRG):
    """Class implmenting DMRG-X [1], whereby local effective energy eigenstates
    are chosen to maximise overlap with the previous step's state, leading to
    convergence on an mid-spectrum eigenstate of the full hamiltonian, as long
    as it is perturbatively close to the original state.

    [1] Khemani, V., Pollmann, F. & Sondhi, S. L. Obtaining Highly Excited
    Eigenstates of Many-Body Localized Hamiltonians by the Density Matrix
    Renormalization Group Approach. Phys. Rev. Lett. 116, 247204 (2016).

    Parameters
    ----------
    ham : MatrixProductOperator
        The hamiltonian in MPO form, should have ~area-law eigenstates.
    p0 : MatrixProductState
        The initial MPS guess, e.g. a computation basis state.
    bond_dims : int or sequence of int
        See :class:`DMRG`.
    cutoffs : float or sequence of float
        See :class:`DMRG`.

    Attributes
    ----------
    k : MatrixProductState
        The current, optimized state.
    energies : list of float
        The list of energies after each sweep.
    """

    def __init__(self, ham, p0, bond_dims, cutoffs=1e-8, bsz=1):
        super().__init__(ham, bond_dims=bond_dims, p0=p0, bsz=bsz,
                         cutoffs=cutoffs)
        # Want to keep track of energy variance as well
        var_ham1 = self.ham.copy()
        var_ham2 = self.ham.copy()
        var_ham1.upper_ind_id = self._k.site_ind_id
        var_ham1.lower_ind_id = "__ham2{}__"
        var_ham2.upper_ind_id = "__ham2{}__"
        var_ham2.lower_ind_id = self._b.site_ind_id
        self.TN_energy2 = self._k | var_ham1 | var_ham2 | self._b
        self.energies.append(self.TN_energy ^ ...)
        self.variances = [(self.TN_energy2 ^ ...) - self.energies[-1]**2]
        self._target_energy = self.energies[-1]

        self.opts = {
            'local_eig_partial_cutoff': 2**11,
            'local_eig_partial_k': 0.02,
            'local_eig_tol': 1e-1,
            'overlap_thresh': 2 / 3,
            'bond_compress_method': 'svd',
            'bond_compress_cutoff_mode': 'sum2',
            'default_sweep_sequence': 'RRLL',
            'bond_expand_rand_strength': 1e-9,
        }

    @property
    def variance(self):
        return self.variances[-1]

    def form_local_ops(self, i, dims, lix, uix):
        self._eff_ham = self.ME_eff_ham()
        self._eff_ovlp = self.ME_eff_ovlp()
        self._eff_ham2 = self.ME_eff_ham2()

        Heff = (self._eff_ham ^ '_HAM')['_HAM'].to_dense(lix, uix)

        return Heff

    def _update_local_state_1site_dmrgx(self, i, direction, **compress_opts):
        """Like ``_update_local_state``, but re-insert all eigenvectors, then
        choose the one with best overlap with ``eff_ovlp``.
        """
        uix, lix = self._k[i].inds, self._b[i].inds
        dims = self._k[i].shape

        # contract remaining hamiltonian and get its dense representation
        Heff = self.form_local_ops(i, dims, lix, uix)

        # eigen-decompose and reshape eigenvectors thus::
        #
        #    |'__ev_ind__'
        #    E
        #   /|\
        #
        D = prod(dims)
        if D <= self.opts['local_eig_partial_cutoff']:
            evals, evecs = eigh(Heff)
        else:
            if isinstance(self.opts['local_eig_partial_k'], float):
                k = int(self.opts['local_eig_partial_k'] * D)
            else:
                k = self.opts['local_eig_partial_k']

            evals, evecs = eigh(
                Heff, sigma=self._target_energy, v0=self._k[i].data,
                k=k, tol=self.opts['local_eig_tol'], backend='scipy')

        evecs = asarray(evecs).reshape(*dims, -1)
        evecs_c = evecs.conj()

        # update tensor at site i with all evecs -> need dummy index
        ki = self._k[i]
        bi = self._b[i]
        ki.modify(data=evecs, inds=(*uix, '__ev_ind__'))

        # find the index of the highest overlap eigenvector, by contracting::
        #
        #           |'__ev_ind__'
        #     o-o-o-E-o-o-o-o-o-o-o
        #     | | | | | | | | | | |
        #     0-0-0-0-0-0-0-0-0-0-0  <- state from previous step
        #
        # choose the eigenvectors with best overlap
        overlaps = np.abs((self._eff_ovlp ^ all).data)

        if self.opts['overlap_thresh'] == 1:
            # just choose the maximum overlap state
            best = np.argmax(overlaps)
        else:
            # else simulteneously reduce energy variance as well
            best_overlaps, = np.where(
                overlaps > np.max(overlaps) * self.opts['overlap_thresh'])

            if len(best_overlaps) == 1:
                # still only one good overlapping eigenvector -> choose that
                best, = best_overlaps
            else:
                # reduce down to the candidate eigenpairs
                evals = evals[best_overlaps]
                evecs = evecs[..., best_overlaps]
                evecs_c = evecs_c[..., best_overlaps]

                # need bra site in place with extra dimension to calc variance
                ki.modify(data=evecs)
                bi.modify(data=evecs_c, inds=(*lix, '__ev_ind__'))

                # now find the variances of the best::
                #
                #           |'__ev_ind__'
                #     o-o-o-E-o-o-o                |'__ev_ind__'  ^2
                #     | | | | | | |          o-o-o-E-o-o-o
                #     H-H-H-H-H-H-H          | | | | | | |
                #     | | | | | | |    -     H-H-H-H-H-H-H
                #     H-H-H-H-H-H-H          | | | | | | |
                #     | | | | | | |          o-o-o-E-o-o-o
                #     o-o-o-E-o-o-o                |'__ev_ind__'
                #           |'__ev_ind__'
                #
                # use einsum notation to get diagonal of left hand term
                en2 = tensor_contract(*self._eff_ham2.tensors,
                                      output_inds=['__ev_ind__']).data

                # then find minimum variance
                best = np.argmin(en2 - evals**2)

        # update site i with the data and drop dummy index too
        ki.modify(data=evecs[..., best], inds=uix)
        bi.modify(data=evecs_c[..., best], inds=lix)
        # store the current effective energy for possibly targeted eigh
        self._target_energy = evals[best]

        tot_en = self._eff_ham ^ all

        self._canonize_after_1site_update(direction, i)

        return evals[best], tot_en

    # def _update_local_state_2site_dmrgx(self, i, direction, **compress_opts):
    #     raise NotImplementedError("2-site DMRGX not implemented yet.")
    #     dims, lix_L, lix_R, lix, uix_L, uix_R, uix, l_bond_ind, u_bond_ind =\
    #         parse_2site_inds_dims(self._k, self._b, i)

    #     # contract remaining hamiltonian and get its dense representation
    #     eff_ham = (self._eff_ham ^ '_HAM')['_HAM']
    #     eff_ham.fuse_((('lower', lix), ('upper', uix)))
    #     A = eff_ham.data

    #     # eigen-decompose and reshape eigenvectors thus::
    #     #
    #     #    ||'__ev_ind__'
    #     #    EE
    #     #   /||\
    #     #
    #     D = prod(dims)
    #     if D <= self.opts['local_eig_partial_cutoff']:
    #         evals, evecs = eigh(A)
    #     else:
    #         if isinstance(self.opts['local_eig_partial_k'], float):
    #             k = int(self.opts['local_eig_partial_k'] * D)
    #         else:
    #             k = self.opts['local_eig_partial_k']

    #         # find the 2-site local state using previous as initial guess
    #         v0 = self._k[i].contract(self._k[i + 1], output_inds=uix).data

    #         evals, evecs = eigh(
    #             A, sigma=self.energies[-1], v0=v0,
    #             k=k, tol=self.opts['local_eig_tol'], backend='scipy')

    def _update_local_state(self, i, **update_opts):
        self.ME_eff_ham.move_to(i)
        self.ME_eff_ham2.move_to(i)
        self.ME_eff_ovlp.move_to(i)

        return {
            1: self._update_local_state_1site_dmrgx,
            # 2: self._update_local_state_2site_dmrgx,
        }[self.bsz](i, **update_opts)

    def sweep(self, direction, canonize=True, verbosity=0, **update_opts):
        """Perform a sweep of the algorithm.

        Parameters
        ----------
        direction : {'R', 'L'}
            Sweep from left to right (->) or right to left (<-) respectively.
        canonize : bool, optional
            Canonize the state first, not needed if doing alternate sweeps.
        verbosity : {0, 1, 2}, optional
            Show a progress bar for the sweep.
        update_opts :
            Supplied to ``self._update_local_state``.
        """
        old_k = self._k.copy().H
        TN_overlap = TensorNetwork([self._k, old_k], virtual=True)

        if canonize:
            {'R': self._k.right_canonize,
             'L': self._k.left_canonize}[direction](bra=self._b)

        direction, begin, sweep = {
            'R': ('right', 'left', range(0, self.n - self.bsz + 1)),
            'L': ('left', 'right', reversed(range(0, self.n - self.bsz + 1))),
        }[direction]

        eff_opts = {'begin': begin, 'bsz': self.bsz, 'cyclic': self.cyclic}
        self.ME_eff_ham = MovingEnvironment(self.TN_energy, **eff_opts)
        self.ME_eff_ham2 = MovingEnvironment(self.TN_energy2, **eff_opts)
        self.ME_eff_ovlp = MovingEnvironment(TN_overlap, **eff_opts)

        if verbosity:
            sweep = progbar(sweep, ncols=80, total=self.n - self.bsz + 1)

        local_ens, tot_ens = zip(*[
            self._update_local_state(i, direction=direction, **update_opts)
            for i in sweep
        ])

        self.local_energies.append(local_ens)
        self.total_energies.append(tot_ens)

        return tot_ens[-1]

    def _compute_post_sweep(self):
        en_var = (self.TN_energy2 ^ ...) - self.energies[-1]**2
        self.variances.append(en_var)

    def _print_post_sweep(self, converged, verbosity=0):
        if verbosity > 1:
            self._k.show()
        if verbosity > 0:
            msg = "Energy={}, Variance={} ... {}"
            msg = msg.format(self.energy, self.variance, "converged!"
                             if converged else "not converged.")
            print(msg, flush=True)

    def _check_convergence(self, tol):
        return self.variance < tol



"""Numpy base linear algebra.
"""

import numpy as np
import numpy.linalg as nla
import scipy.linalg as scla

import quimb as qu
from .autoblock import eigensystem_autoblocked


_NUMPY_EIG_FUNCS = {
    (True, True): nla.eigh,
    (True, False): nla.eig,
    (False, True): nla.eigvalsh,
    (False, False): nla.eigvals,
}


def eig_numpy(A, sort=True, isherm=True, return_vecs=True, autoblock=False):
    """Numpy based dense eigensolve.

    Parameters
    ----------
    A : array_like
        The operator to decompose.
    sort : bool, optional
        Whether to sort into ascending order.
    isherm : bool, optional
        Whether ``A`` is hermitian.
    return_vecs : bool, optional
        Whether to return the eigenvectors.
    autoblock : bool, optional
        If true, automatically identify and exploit symmetries appearing in the
        current basis as block diagonals formed via permutation of rows and
        columns.

    Returns
    -------
    evals : 1D-array
        The eigenvalues.
    evecs : qarray
        If ``return_vecs=True``, the eigenvectors.
    """
    if autoblock:
        return eigensystem_autoblocked(A, sort=sort, isherm=isherm,
                                       return_vecs=return_vecs)

    evals = _NUMPY_EIG_FUNCS[return_vecs, isherm](A)

    if return_vecs:
        evals, evecs = evals

        if sort:
            sortinds = np.argsort(evals)
            evals, evecs = evals[sortinds], evecs[:, sortinds]

        return evals, qu.qarray(evecs)

    if sort:
        evals.sort()

    return evals


def sort_inds(a, method, sigma=None):
    """Return the sorting inds of a list

    Parameters
    ----------
        a : array_like
            List to base sort on.
        method : str
            Method of sorting list, one of
                * "LM" - Largest magnitude first
                * "SM" - Smallest magnitude first
                * "SA" - Smallest algebraic first
                * "SR" - Smallest real part first
                * "SI" - Smallest imaginary part first
                * "LA" - Largest algebraic first
                * "LR" - Largest real part first
                * "LI" - Largest imaginary part first
                * "TM" - Magnitude closest to target sigma first
                * "TR" - Real part closest to target sigma first
                * "TI" - Imaginary part closest to target sigma first
        sigma : float, optional
            The target if method={"TM", "TR", or "TI"}.

    Returns
    -------
        inds : array of int
            Indices that would sort `a` based on `method`
    """
    _SORT_FUNCS = {
        "LM": lambda a: -abs(a),
        "SM": lambda a: -abs(1 / a),
        "SA": lambda a: a,
        "SR": lambda a: a.real,
        "SI": lambda a: a.imag,
        "LA": lambda a: -a,
        "LR": lambda a: -a.real,
        "LI": lambda a: -a.imag,
        "TM": lambda a: -1 / abs(abs(a) - sigma),
        "TR": lambda a: -1 / abs(a.real - sigma),
        "TI": lambda a: -1 / abs(a.imag - sigma),
    }
    return np.argsort(_SORT_FUNCS[method.upper()](a))


_DENSE_EIG_METHODS = {
    (True, True, False): nla.eigh,
    (True, False, False): nla.eigvalsh,
    (False, True, False): nla.eig,
    (False, False, False): nla.eigvals,
    (True, True, True): scla.eigh,
    (True, False, True): scla.eigvalsh,
    (False, True, True): scla.eig,
    (False, False, True): scla.eigvals,
}


def eigs_numpy(A, k, B=None, which=None, return_vecs=True,
               sigma=None, isherm=True, P=None, sort=True, **eig_opts):
    """Partial eigen-decomposition using numpy's dense linear algebra.

    Parameters
    ----------
    A : array_like or quimb.Lazy
        Operator to partially eigen-decompose.
    k : int
        Number of eigenpairs to return.
    B : array_like or quimb.Lazy
        If given, the RHS operator defining a generalized eigen problem.
    which : str, optional
        Which part of the spectrum to target.
    return_vecs : bool, optional
        Whether to return eigenvectors.
    sigma : None or float, optional
        Target eigenvalue.
    isherm : bool, optional
        Whether `a` is hermitian.
    P : array_like or quimb.Lazy
        Perform the eigensolve in the subspace defined by this projector.
    sort : bool, optional
        Whether to sort reduced list of eigenpairs into ascending order.
    eig_opts
        Settings to pass to numpy.eig... functions.

    Returns
    -------
        lk, (vk): k eigenvalues (and eigenvectors) sorted according to which
    """
    if isinstance(A, qu.Lazy):
        A = A()
    if isinstance(B, qu.Lazy):
        B = B()
    if isinstance(P, qu.Lazy):
        P = P()

    # project into subspace
    if P is not None:
        A = qu.dag(P) @ (A @ P)

    generalized = B is not None

    eig_fn = _DENSE_EIG_METHODS[(isherm, return_vecs, generalized)]

    if generalized:
        eig_opts['b'] = B

    # these might be given for partial eigsys but not relevant for numpy
    eig_opts.pop('ncv', None)
    eig_opts.pop('v0', None)
    eig_opts.pop('tol', None)
    eig_opts.pop('maxiter', None)
    eig_opts.pop('EPSType', None)

    if return_vecs:
        # get all eigenpairs
        lk, vk = eig_fn(A.A if qu.issparse(A) else A, **eig_opts)

        # sort and trim according to which k we want
        sk = sort_inds(lk, method=which, sigma=sigma)[:k]
        lk, vk = lk[sk], vk[:, sk]

        # also potentially sort into ascending order
        if sort:
            so = np.argsort(lk)
            lk, vk = lk[so], vk[:, so]

        # map eigenvectors out of subspace
        if P is not None:
            vk = P @ vk

        return lk, qu.qarray(vk)

    else:
        # get all eigenvalues
        lk = eig_fn(A.A if qu.issparse(A) else A, **eig_opts)

        # sort and trim according to which k we want
        sk = sort_inds(lk, method=which, sigma=sigma)[:k]
        lk = lk[sk]

        # also potentially sort into ascending order
        return np.sort(lk) if sort else lk


def svds_numpy(a, k, return_vecs=True, **_):
    """Partial singular value decomposition using numpys (full) singular value
    decomposition.

    Parameters
    ----------
    a : array_like
        Operator to decompose.
    k : int, optional
        Number of singular value triplets to retrieve.
    return_vecs : bool, optional
        whether to return the computed vecs or values only

    Returns
    -------
    (uk,) sk (, vkt) :
        Singlar value triplets.
    """
    if return_vecs:
        uk, sk, vkt = nla.svd(a.A if qu.issparse(a) else a, compute_uv=True)
        return qu.qarray(uk[:, :k]), sk[:k], qu.qarray(vkt[:k, :])
    else:
        sk = nla.svd(a.A if qu.issparse(a) else a, compute_uv=False)
        return sk[:k]



"""Support for optimizing tensor networks using ``autograd`` or ``jax`` to
automatically derive gradients for input to scipy.
"""
import re
import functools
import importlib

import tqdm
import numpy as np
from autoray import do

from .tensor_core import contract_backend, Tensor, TensorNetwork, PTensor
from .array_ops import iscomplex
from ..core import qarray

if importlib.util.find_spec("jax") is not None:
    _DEFAULT_BACKEND = 'jax'
else:
    _DEFAULT_BACKEND = 'autograd'


_REAL_CONVERSION = {
    'float32': 'float32',
    'float64': 'float64',
    'complex64': 'float32',
    'complex128': 'float64',
}

_COMPLEX_CONVERSION = {
    'float32': 'complex64',
    'float64': 'complex128',
    'complex64': 'complex64',
    'complex128': 'complex128',
}


def equivalent_real_type(x):
    return _REAL_CONVERSION[x.dtype.name]


def equivalent_complex_type(x):
    return _COMPLEX_CONVERSION[x.dtype.name]


class Vectorizer:
    """Object for mapping a sequence of mixed real/complex n-dimensional arrays
    to a single numpy vector and back and forth.
    """

    def __init__(self, arrays):
        self.shapes = [x.shape for x in arrays]
        self.iscomplexes = [iscomplex(x) for x in arrays]
        self.dtypes = [x.dtype for x in arrays]
        self.sizes = [np.prod(s) for s in self.shapes]
        self.d = sum(
            (1 + int(cmplx)) * size
            for size, cmplx in zip(self.sizes, self.iscomplexes)
        )
        self.pack(arrays)

    def pack(self, arrays, name='vector'):

        # scipy's optimization routines require real, double data
        if not hasattr(self, name):
            setattr(self, name, np.empty(self.d, 'float64'))
        x = getattr(self, name)

        i = 0
        for array, size, cmplx in zip(arrays, self.sizes, self.iscomplexes):
            array = np.asarray(array)
            if not cmplx:
                x[i:i + size] = array.reshape(-1)
                i += size
            else:
                real_view = array.reshape(-1).view(equivalent_real_type(array))
                x[i:i + 2 * size] = real_view
                i += 2 * size

        return x

    def unpack(self, vector=None, conj=False):
        """Turn the single, flat ``vector`` into a sequence of arrays.
        """
        if vector is None:
            vector = self.vector

        i = 0
        arrays = []
        for shape, size, cmplx, dtype in zip(self.shapes, self.sizes,
                                             self.iscomplexes, self.dtypes):
            if not cmplx:
                array = vector[i:i + size]
                array.shape = shape
                i += size
            else:
                array = vector[i:i + 2 * size]
                array = array.view(equivalent_complex_type(array))
                # XXX: need to conjugate for autograd/jax, convention??
                if conj:
                    array = np.conj(array)
                array.shape = shape
                i += 2 * size

            if array.dtype != dtype:
                array = array.astype(dtype)
            arrays.append(array)

        return arrays


def parse_network_to_ag(tn, constant_tags, backend=_DEFAULT_BACKEND):
    tn_ag = tn.copy()
    variables = []

    variable_tag = "__VARIABLE{}__"

    for t in tn_ag:
        # check if tensor has any of the constant tags
        if t.tags & constant_tags:
            t.modify(data=constant(t.data, backend=backend))
            continue

        if isinstance(t, PTensor):
            data = t.params
        else:
            data = t.data

        # jax doesn't like numpy.ndarray subclasses...
        if isinstance(data, qarray):
            data = data.A

        # append the raw data but mark the corresponding tensor for reinsertion
        variables.append(data)
        t.add_tag(variable_tag.format(len(variables) - 1))

    return tn_ag, variables


def constant(x, backend=_DEFAULT_BACKEND):
    if backend == 'jax' and isinstance(x, qarray):
        x = x.A
    return do('array', x, like=backend)


def constant_t(t, backend=_DEFAULT_BACKEND):
    ag_t = t.copy()
    ag_t.modify(data=constant(ag_t.data, backend=backend))
    return ag_t


def constant_tn(tn, backend=_DEFAULT_BACKEND):
    """Convert a tensor network's arrays to constants.
    """
    ag_tn = tn.copy()
    ag_tn.apply_to_arrays(functools.partial(constant, backend=backend))
    return ag_tn


def convert_args_to_jax(fn):

    @functools.wraps(fn)
    def jax_args_fn(arrays):
        jarrays = [constant(array) for array in arrays]
        return fn(jarrays)

    return jax_args_fn


variable_finder = re.compile(r'__VARIABLE(\d+)__')


def inject_(arrays, tn):
    for t in tn:
        for tag in t.tags:
            match = variable_finder.match(tag)
            if match is not None:
                i = int(match.groups(1)[0])

                if isinstance(t, PTensor):
                    t.params = arrays[i]
                else:
                    t.modify(data=arrays[i])

                break


class TNOptimizer:

    def __init__(
        self,
        tn,
        loss_fn,
        norm_fn=None,
        loss_constants=None,
        loss_kwargs=None,
        constant_tags=None,
        loss_target=None,
        optimizer='L-BFGS-B',
        progbar=True,
        autograd_backend='AUTO',
        # the rest are ignored for compat
        learning_rate=None,
        learning_decay_steps=None,
        learning_decay_rate=None,
    ):
        self.progbar = progbar
        self.optimizer = optimizer
        self.constant_tags = (
            set() if constant_tags is None else set(constant_tags)
        )

        if autograd_backend.upper() == 'AUTO':
            autograd_backend = _DEFAULT_BACKEND

        self.autograd_backend = autograd_backend

        # use identity if no nomalization required
        if norm_fn is None:
            def norm_fn(x):
                return x

        self.norm_fn = norm_fn

        self.loss_kwargs = {} if loss_kwargs is None else dict(loss_kwargs)

        self.loss_constants = {}
        if loss_constants is not None:
            for k, v in loss_constants.items():
                # check if tensor network supplied
                if isinstance(v, TensorNetwork):
                    # convert it to constant TN
                    self.loss_constants[k] = constant_tn(v, autograd_backend)
                elif isinstance(v, Tensor):
                    self.loss_constants[k] = constant_t(v, autograd_backend)
                else:
                    self.loss_constants[k] = constant(v, autograd_backend)

        self.loss_fn = functools.partial(
            loss_fn, **self.loss_constants, **self.loss_kwargs
        )

        self.loss = np.inf
        self.loss_best = np.inf
        self.loss_target = loss_target
        self._n = 0

        self.tn_opt, self.variables = parse_network_to_ag(
            tn, self.constant_tags
        )

        # this handles storing and packing /  unpacking many arrays as a vector
        self.vctrzr = Vectorizer(self.variables)

        def func(arrays):
            # need to set backend explicitly as mixing with numpy arrays
            with contract_backend(self.autograd_backend):
                inject_(arrays, self.tn_opt)
                return self.loss_fn(self.norm_fn(self.tn_opt))

        if self.autograd_backend == 'jax':
            import jax
            self.func_jit = jax.jit(func)
            self.grad_jit = jax.jit(jax.grad(self.func_jit))
        elif self.autograd_backend == 'autograd':
            import autograd
            self.func_jit = func
            self.grad_jit = autograd.grad(func)

        def vectorized_func(x):
            arrays = self.vctrzr.unpack(x, conj=True)
            ag_result = self.func_jit(arrays)
            self.loss = np.asarray(ag_result)
            self._n += 1
            return self.loss

        def vectorized_grad(x):
            arrays = self.vctrzr.unpack(x, conj=True)
            ag_grads = self.grad_jit(arrays)
            vec_grad = self.vctrzr.pack(ag_grads, 'grad')
            return vec_grad

        self.vectorized_func = vectorized_func
        self.vectorized_grad = vectorized_grad

    @property
    def nevals(self):
        """The number of gradient evaluations.
        """
        return self._n

    def inject_res_vector_and_return_tn(self):
        inject_(self.vctrzr.unpack(self.res.x, conj=True), self.tn_opt)
        self.vctrzr.vector[:] = self.res.x
        return self.norm_fn(self.tn_opt)

    def optimize(self, n, tol=None, **options):
        from scipy.optimize import minimize

        try:
            pbar = tqdm.tqdm(total=n, disable=not self.progbar)

            def callback(_):
                pbar.update()
                pbar.set_description(f"{self.loss}")

            self.res = minimize(
                fun=self.vectorized_func,
                jac=self.vectorized_grad,
                x0=self.vctrzr.vector,
                callback=callback,
                tol=tol,
                method=self.optimizer,
                options=dict(
                    maxiter=n,
                    **options,
                )
            )

        finally:
            pbar.close()

        return self.inject_res_vector_and_return_tn()

    def optimize_basinhopping(self, n, nhop, temperature=1.0):
        from scipy.optimize import basinhopping

        try:
            pbar = tqdm.tqdm(total=n * nhop, disable=not self.progbar)

            def callback(x, f, accept):
                if self.loss_target is not None:
                    if self.loss_best < self.loss_target:
                        return True

            def inner_callback(xk):
                self.loss_best = min(self.loss_best, self.loss)
                pbar.update()
                msg = f"{self.loss} [best: {self.loss_best}] "
                pbar.set_description(msg)

            self.res = basinhopping(
                func=self.vectorized_func,
                x0=self.vctrzr.vector,
                niter=nhop,
                minimizer_kwargs=dict(
                    jac=self.vectorized_grad,
                    method=self.optimizer,
                    callback=inner_callback,
                    options=dict(
                        maxiter=n,
                    )
                ),
                callback=callback,
                T=temperature,
            )

        finally:
            pbar.close()

        return self.inject_res_vector_and_return_tn()
